# Generated from /Users/pelson/dev/scitools/cf-units/cf_units/_udunits2_parser/udunits2Parser.g4 by ANTLR 4.7.2
from antlr4 import *

if __name__ is not None and "." in __name__:
    from .udunits2Parser import udunits2Parser
else:
    from udunits2Parser import udunits2Parser

# This class defines a complete generic visitor for a parse tree produced by udunits2Parser.


class udunits2ParserVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by udunits2Parser#unit_spec.
    def visitUnit_spec(self, ctx: udunits2Parser.Unit_specContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by udunits2Parser#shift_spec.
    def visitShift_spec(self, ctx: udunits2Parser.Shift_specContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by udunits2Parser#product.
    def visitProduct(self, ctx: udunits2Parser.ProductContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by udunits2Parser#power.
    def visitPower(self, ctx: udunits2Parser.PowerContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by udunits2Parser#basic_spec.
    def visitBasic_spec(self, ctx: udunits2Parser.Basic_specContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by udunits2Parser#integer.
    def visitInteger(self, ctx: udunits2Parser.IntegerContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by udunits2Parser#number.
    def visitNumber(self, ctx: udunits2Parser.NumberContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by udunits2Parser#timestamp.
    def visitTimestamp(self, ctx: udunits2Parser.TimestampContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by udunits2Parser#signed_clock.
    def visitSigned_clock(self, ctx: udunits2Parser.Signed_clockContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by udunits2Parser#timezone_offset.
    def visitTimezone_offset(self, ctx: udunits2Parser.Timezone_offsetContext):
        return self.visitChildren(ctx)


del udunits2Parser



# Copyright cf-units contributors
#
# This file is part of cf-units and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

import cf_units._udunits2_parser.graph as g
from cf_units._udunits2_parser import parse


def test_Node_attributes():
    n = g.Node(a=1, kwarg="two", arbitrary_kwargs=3)

    assert n.a == 1
    assert n.kwarg == "two"
    assert n.arbitrary_kwargs == 3


def test_Node_str():
    n = g.Node(a=1, kwarg="two", arbitrary_kwargs=3)
    assert str(n) == "Node(a=1, kwarg='two', arbitrary_kwargs=3)"


def test_Node_children():
    n = g.Node(a=1, kwarg="two", arbitrary_kwargs=3)
    # Ordered, and consistent.
    assert n.children() == [1, "two", 3]


def test_large_graph():
    graph = parse("m2/4.1.2π per second @ 10")
    assert isinstance(graph, g.Shift)

    unit, shift_from = graph.children()
    assert isinstance(shift_from, g.Number)
    assert str(shift_from) == "10"

    assert isinstance(unit, g.Divide)
    lhs, rhs = unit.children()
    assert str(lhs) == "m^2/4.1·.2·π"
    assert str(rhs) == "second"

    assert isinstance(lhs, g.Multiply)
    lhs, rhs = lhs.children()
    assert str(lhs) == "m^2/4.1·.2"
    assert str(rhs) == "π"

    assert isinstance(lhs, g.Multiply)
    lhs, rhs = lhs.children()
    assert str(lhs) == "m^2/4.1"
    assert str(rhs) == ".2"

    assert isinstance(lhs, g.Divide)
    lhs, rhs = lhs.children()
    assert str(lhs) == "m^2"
    assert str(rhs) == "4.1"

    assert isinstance(lhs, g.Raise)
    lhs, rhs = lhs.children()
    assert str(lhs) == "m"
    assert str(rhs) == "2"



# Copyright cf-units contributors
#
# This file is part of cf-units and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test function :func:`cf_units.num2date`."""

import unittest

from cf_units import num2date


class Test(unittest.TestCase):
    def test_num2date_wrong_calendar(self):
        with self.assertRaisesRegex(
            ValueError, "illegal calendar or reference date"
        ):
            num2date(
                1,
                "days since 1970-01-01",
                calendar="360_day",
                only_use_cftime_datetimes=False,
                only_use_python_datetimes=True,
            )


if __name__ == "__main__":
    unittest.main()



# Copyright cf-units contributors
#
# This file is part of cf-units and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

import cf_units._udunits2_parser.graph as graph  # noqa: E402
from cf_units._udunits2_parser import parse as _parse  # noqa: E402


class TeXVisitor(graph.Visitor):
    def _format(self, fmt, lhs, rhs):
        return fmt.format(self.visit(lhs), self.visit(rhs))

    def visit_Identifier(self, node):
        token = str(node)
        if token.startswith("micro"):
            token = token.replace("micro", r"{\mu}")
        return token

    def visit_Raise(self, node):
        return self._format("{{{}}}^{{{}}}", node.lhs, node.rhs)

    def visit_Multiply(self, node):
        return self._format(r"{{{}}}\cdot{{{}}}", node.lhs, node.rhs)

    def visit_Divide(self, node):
        return self._format(r"\frac{{{}}}{{{}}}", node.lhs, node.rhs)

    def visit_Shift(self, node):
        return self._format("{{{}}} @ {{{}}}", node.unit, node.shift_from)

    def generic_visit(self, node):
        result = [self.visit(child) for child in node.children()]
        if not result:
            result = str(node)
        return result


def tex(unit_str):
    tree = _parse(unit_str)
    return TeXVisitor().visit(tree)



# Copyright cf-units contributors
#
# This file is part of cf-units and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test function :func:`cf_units.date2num`."""

import datetime
import unittest

import numpy as np

from cf_units import date2num


class Test(unittest.TestCase):
    def setUp(self):
        self.unit = "seconds since 1970-01-01"
        self.calendar = "gregorian"

    def test_single(self):
        date = datetime.datetime(1970, 1, 1, 0, 0, 5)
        exp = 5.0
        res = date2num(date, self.unit, self.calendar)
        # num2date won't return an exact value representing the date,
        # even if one exists
        self.assertAlmostEqual(exp, res, places=4)

    def test_sequence(self):
        dates = [
            datetime.datetime(1970, 1, 1, 0, 0, 20),
            datetime.datetime(1970, 1, 1, 0, 0, 40),
            datetime.datetime(1970, 1, 1, 0, 1),
            datetime.datetime(1970, 1, 1, 0, 1, 20),
            datetime.datetime(1970, 1, 1, 0, 1, 40),
        ]
        exp = [20.0, 40.0, 60.0, 80, 100.0]
        res = date2num(dates, self.unit, self.calendar)
        np.testing.assert_array_almost_equal(exp, res, decimal=4)

    def test_multidim_sequence(self):
        dates = [
            [
                datetime.datetime(1970, 1, 1, 0, 0, 20),
                datetime.datetime(1970, 1, 1, 0, 0, 40),
                datetime.datetime(1970, 1, 1, 0, 1),
            ],
            [
                datetime.datetime(1970, 1, 1, 0, 1, 20),
                datetime.datetime(1970, 1, 1, 0, 1, 40),
                datetime.datetime(1970, 1, 1, 0, 2),
            ],
        ]
        exp_shape = (2, 3)
        res = date2num(dates, self.unit, self.calendar)
        self.assertEqual(exp_shape, res.shape)

    def test_discard_mircosecond(self):
        date = datetime.datetime(1970, 1, 1, 0, 0, 5, 750000)
        exp = 5.0
        res = date2num(date, self.unit, self.calendar)

        self.assertAlmostEqual(exp, res, places=4)

    def test_long_time_interval(self):
        # This test should fail with an error that we need to catch properly.
        unit = "years since 1970-01-01"
        date = datetime.datetime(1970, 1, 1, 0, 0, 5)
        exp_emsg = 'interval of "months", "years" .* got "years".'
        with self.assertRaisesRegex(ValueError, exp_emsg):
            date2num(date, unit, self.calendar)


if __name__ == "__main__":
    unittest.main()



# Copyright cf-units contributors
#
# This file is part of cf-units and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Units of measure.

Wrapper for Unidata/UCAR UDUNITS-2.

See also: `UDUNITS-2
<http://www.unidata.ucar.edu/software/udunits/udunits-2/udunits2.html>`_.

"""

# cython: nonecheck=True

import numpy as np

cimport numpy as np
from libc cimport errno, string


cdef int _STRING_BUFFER_DEPTH = 128
cdef list _UT_STATUS = ['UT_SUCCESS', 'UT_BAD_ARG', 'UT_EXISTS', 'UT_NO_UNIT',
                        'UT_OS', 'UT_NOT_SAME_SYSTEM', 'UT_MEANINGLESS',
                        'UT_NO_SECOND', 'UT_VISIT_ERROR', 'UT_CANT_FORMAT',
                        'UT_SYNTAX', 'UT_UNKNOWN', 'UT_OPEN_ARG',
                        'UT_OPEN_ENV', 'UT_OPEN_DEFAULT', 'UT_PARSE']

UT_ASCII = 0
UT_ISO_8859_1 = 1
UT_LATIN1 = UT_ISO_8859_1
UT_UTF8 = 2

UT_NAMES = 4
UT_DEFINITION = 8


##### Wrapper classes #####

cdef class System:
    """
    Wrapper class for UDUNITS-2 ut_system. Calls ut_free_system() on
    deallocation.

    """
    cdef ut_system *csystem

    def __cinit__(self):
        self.csystem = NULL

    def __dealloc__(self):
        ut_free_system(self.csystem)


cdef class Unit:
    """
    Wrapper class for UDUNITS-2 ut_unit. Calls ut_free() on
    deallocation.

    """
    cdef ut_unit *cunit
    cdef System system

    def __cinit__(self):
        self.cunit = NULL
        self.system = None

    def __dealloc__(self):
        ut_free(self.cunit)


cdef class Converter:
    """
    Wrapper class for UDUNITS-2 cv_converter. Calls cv_free() on
    deallocation.

    """
    cdef cv_converter *cconverter

    def __cinit__(self):
        self.cconverter = NULL

    def __dealloc__(self):
        cv_free(self.cconverter)


cdef class ErrorMessageHandler:
    """
    Wrapper class for UDUNITS-2 ut_error_message_handler typedef.

    """
    cdef ut_error_message_handler chandler

    def __cinit__(self):
        self.chandler = NULL

# The following wrap_* functions should be called directly after the api call
# which returned their arguments, in order that in the event of an error the
# correct ut_status and errno values are retreived by _raise_error()

cdef System wrap_system(ut_system* csystem):
    if csystem is NULL:
        _raise_error()
    cdef System sytem = System()
    sytem.csystem = csystem
    return sytem

cdef Unit wrap_unit(System system, ut_unit* cunit):
    if cunit is NULL:
        _raise_error()
    cdef Unit unit = Unit()
    unit.cunit = cunit
    unit.system = system
    return unit

cdef Converter wrap_converter(cv_converter* cconverter):
    if cconverter is NULL:
        _raise_error()
    cdef Converter converter = Converter()
    converter.cconverter = cconverter
    return converter

cdef ErrorMessageHandler _ignore = ErrorMessageHandler()
_ignore.chandler = ut_ignore
ignore = _ignore

# Convenience object to avoid having to do None handling either here or in
# user code
NULL_UNIT = Unit()


##### Exception class #####

class UdunitsError(Exception):
    """
    UDUNITS-2 call resulted in an error.

    Attributes:

    * status:
        The UDUNITS-2 ut_status value which resulted from the error.
    * errnum:
        The errno value which resulted from the error.

    """
    def __init__(self, ut_status status, int errnum):
        self.status = status
        self.errnum = errnum

    def status_msg(self):
        """
        String representation of the UDUNITS-2 ut_status value which resulted
        from the error.

        """
        if 0 <= self.status < len(_UT_STATUS):
            return _UT_STATUS[self.status]
        else:
            return 'UNKNOWN'

    def error_msg(self):
        """
        The message string associated with the errno value which resulted from
        the error.

        """
        return string.strerror(self.errnum) if self.errnum else ''

    def __str__(self):
        str_err = ': {}'.format(string.strerror(self.errnum)) \
                  if self.errnum else ''
        return '{}{}'.format(self.status_msg(), str_err)


def _raise_error():
    """
    Raise a UdunitsError with the current value of errno and ut_status

    """
    errnum = errno.errno
    errno.errno = 0
    status = ut_get_status()
    raise UdunitsError(status, errnum)


##### Wrapper functions #####

# The following functions wrap the corresponding ut_* or cv_* api calls.
# See the UDUNITS-2 documentation for details.

def read_xml(char* path=NULL):
    cdef ut_system* csystem = ut_read_xml(path)
    return wrap_system(csystem)

def get_unit_by_name(System system, char* name):
    cdef ut_unit* cunit = ut_get_unit_by_name(system.csystem, name)
    return wrap_unit(system, cunit)

def clone(Unit unit):
    cdef ut_unit* cunit = ut_clone(unit.cunit)
    return wrap_unit(unit.system, cunit)

def is_dimensionless(Unit unit):
    return <bint>ut_is_dimensionless(unit.cunit)

def compare(Unit unit1, Unit unit2):
    return ut_compare(unit1.cunit, unit2.cunit)

def are_convertible(Unit unit1, Unit unit2):
    return <bint>ut_are_convertible(unit1.cunit, unit2.cunit)

def get_converter(Unit fr, Unit to):
    cdef cv_converter* cconverter = ut_get_converter(fr.cunit, to.cunit)
    return wrap_converter(cconverter)

def scale(double factor, Unit unit):
    cdef ut_unit* cunit = ut_scale(factor, unit.cunit)
    return wrap_unit(unit.system, cunit)

def offset(Unit unit, double offset):
    cdef ut_unit* cunit = ut_offset(unit.cunit, offset)
    return wrap_unit(unit.system, cunit)

def offset_by_time(Unit unit, double origin):
    cdef ut_unit* cunit = ut_offset_by_time(unit.cunit, origin)
    return wrap_unit(unit.system, cunit)

def multiply(Unit unit1, Unit unit2):
    cdef ut_unit* cunit = ut_multiply(unit1.cunit, unit2.cunit)
    return wrap_unit(unit1.system, cunit)

def invert(Unit unit):
    cdef ut_unit* cunit =  ut_invert(unit.cunit)
    return wrap_unit(unit.system, cunit)

def divide(Unit numer, Unit denom):
    cdef ut_unit* cunit = ut_divide(numer.cunit, denom.cunit)
    return wrap_unit(numer.system, cunit)

def raise_(Unit unit, int power):
    cdef ut_unit* cunit = ut_raise(unit.cunit, power)
    return wrap_unit(unit.system, cunit)

def root(Unit unit, int root):
    cdef ut_unit* cunit = ut_root(unit.cunit, root)
    return wrap_unit(unit.system, cunit)

def log(double base, Unit reference):
    cdef ut_unit* cunit = ut_log(base, reference.cunit)
    return wrap_unit(reference.system, cunit)


def parse(System system, char* string, ut_encoding encoding):
    cdef ut_unit* cunit = ut_parse(system.csystem, string, encoding)
    return wrap_unit(system, cunit)

def format(Unit unit, unsigned opts=0):
    cdef bytearray buf = bytearray(_STRING_BUFFER_DEPTH)
    n = ut_format(unit.cunit, buf, len(buf), opts)
    if n >= _STRING_BUFFER_DEPTH:
        buf = bytearray(n + 1)
        n = ut_format(unit.cunit, buf, len(buf), opts)
    if n == -1:
        _raise_error()
    return bytes(buf[:n])

def encode_date(int year, int month, int day):
    return ut_encode_date(year, month, day)

def encode_clock(int hours, int minutes, double seconds):
    return ut_encode_clock(hours, minutes, seconds)

def encode_time(int year, int month, int day,
                int hour, int minute, double second):
    return ut_encode_time(year, month, day, hour, minute, second)

def decode_time(double value):
    cdef int year, month, day, hour, minute
    cdef double second, resolution
    ut_decode_time(value, &year, &month, &day, &hour, &minute, &second,
                   &resolution)
    return (year, month, day, hour, minute, second, resolution)

def set_error_message_handler(ErrorMessageHandler handler):
    cdef ErrorMessageHandler result = ErrorMessageHandler()
    result.chandler = ut_set_error_message_handler(handler.chandler)
    return result

def convert_float(Converter converter, float value):
    return cv_convert_float(converter.cconverter, value)

def convert_floats(Converter converter, np.ndarray[np.float32_t] in_, np.ndarray[np.float32_t] out):
    cv_convert_floats(converter.cconverter, <float*> in_.data, in_.size, <float*> out.data)
    return out

def convert_double(Converter converter, double value):
    return cv_convert_double(converter.cconverter, value)

def convert_doubles(Converter converter, np.ndarray[np.float64_t] in_, np.ndarray[np.float64_t] out):
    cv_convert_doubles(converter.cconverter, <double*> in_.data, in_.size, <double*> out.data)
    return out



# Copyright cf-units contributors
#
# This file is part of cf-units and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `cf_units._discard_microsecond` function."""

import datetime
import unittest

import cftime
import numpy as np
import numpy.ma as ma

from cf_units import _discard_microsecond as discard_microsecond


class Test__datetime(unittest.TestCase):
    def setUp(self):
        self.kwargs = dict(year=1, month=2, day=3, hour=4, minute=5, second=6)
        self.expected = datetime.datetime(**self.kwargs, microsecond=0)

    def test_single(self):
        dt = datetime.datetime(**self.kwargs, microsecond=7)
        actual = discard_microsecond(dt)
        self.assertEqual(self.expected, actual)

    def test_multi(self):
        shape = (5, 2)
        n = np.prod(shape)

        dates = np.array(
            [datetime.datetime(**self.kwargs, microsecond=i) for i in range(n)]
        ).reshape(shape)
        actual = discard_microsecond(dates)
        expected = np.array([self.expected] * n).reshape(shape)
        np.testing.assert_array_equal(expected, actual)


class Test__cftime(unittest.TestCase):
    def setUp(self):
        self.kwargs = dict(year=1, month=2, day=3, hour=4, minute=5, second=6)
        self.calendars = cftime._cftime._calendars

    def test_single(self):
        for calendar in self.calendars:
            dt = cftime.datetime(**self.kwargs, calendar=calendar)
            actual = discard_microsecond(dt)
            expected = cftime.datetime(
                **self.kwargs, microsecond=0, calendar=calendar
            )
            self.assertEqual(expected, actual)

    def test_multi(self):
        shape = (2, 5)
        n = np.prod(shape)

        for calendar in self.calendars:
            dates = np.array(
                [
                    cftime.datetime(**self.kwargs, calendar=calendar)
                    for i in range(n)
                ]
            ).reshape(shape)
            actual = discard_microsecond(dates)
            expected = np.array(
                [
                    cftime.datetime(
                        **self.kwargs, microsecond=0, calendar=calendar
                    )
                ]
                * n
            ).reshape(shape)
            np.testing.assert_array_equal(expected, actual)


class Test__falsy(unittest.TestCase):
    def setUp(self):
        self.kwargs = dict(year=1, month=2, day=3, hour=4, minute=5, second=6)
        self.calendar = "360_day"
        microsecond = 7
        self.cftime = cftime.datetime(
            **self.kwargs, microsecond=microsecond, calendar=self.calendar
        )
        self.datetime = datetime.datetime(
            **self.kwargs, microsecond=microsecond
        )

    def test_single__none(self):
        self.assertIsNone(discard_microsecond(None))

    def test_single__false(self):
        self.assertFalse(discard_microsecond(False))

    def test_multi__falsy(self):
        falsy = np.array([None, False, 0])
        actual = discard_microsecond(falsy)
        np.testing.assert_array_equal(falsy, actual)

    def test_masked(self):
        data = [self.cftime, self.datetime, self.cftime, self.datetime]
        mask = [1, 0, 0, 1]
        dates = ma.masked_array(data, mask=mask)
        actual = discard_microsecond(dates)
        expected = np.array(
            [
                ma.masked,
                datetime.datetime(**self.kwargs),
                cftime.datetime(**self.kwargs, calendar=self.calendar),
                ma.masked,
            ]
        )
        self.assertEqual(expected.shape, actual.shape)
        for i, masked in enumerate(mask):
            if masked:
                self.assertIs(expected[i], actual[i])
            else:
                self.assertEqual(expected[i], actual[i])


if __name__ == "__main__":
    unittest.main()



# Copyright cf-units contributors
#
# This file is part of cf-units and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `cf_units.as_unit` function."""

import copy
import unittest

from cf_units import Unit, as_unit


class StubUnit(object):
    def __init__(self, calendar=None):
        self.calendar = str(calendar) if calendar is not None else None
        # udunit category
        self.category = 2
        # ut_unit should be an integer but not under test
        self.ut_unit = 0
        self.origin = "hours since 1970-01-01 00:00:00"

    def __str__(self):
        return self.origin


class TestAll(unittest.TestCase):
    def _assert_unit_equal(self, unit1, unit2):
        # Custom equality assertion of units since equality of the Unit class
        # utilises as_unit.
        for attribute in ["origin", "calendar", "category"]:
            self.assertEqual(
                getattr(unit1, attribute), getattr(unit2, attribute)
            )

    def test_cf_unit(self):
        # Ensure that as_unit returns the same cf_unit.Unit object and
        # remains unchanged.
        unit = Unit("hours since 1970-01-01 00:00:00", calendar="360_day")
        target = copy.copy(unit)
        result = as_unit(unit)

        self._assert_unit_equal(result, target)
        self.assertIs(result, unit)

    def test_non_cf_unit_no_calendar(self):
        # On passing as_unit a cf_unit.Unit-like object (not a cf_unit.Unit
        # object) with no calendar, ensure that a cf_unit.Unit is returned
        # with default calendar (Gregorian).
        unit = StubUnit()
        result = as_unit(unit)

        self.assertEqual(result.calendar, "gregorian")
        self.assertIsInstance(result, Unit)

    def test_non_cf_unit_with_calendar(self):
        # On passing as_unit a cf_unit.Unit-like object (not a cf_unit.Unit
        # object) with calendar, ensure that a cf_unit.Unit is returned
        # with the same calendar specified in the original unit-like object.
        unit = StubUnit("360_day")
        target = copy.copy(unit)
        result = as_unit(unit)
        self._assert_unit_equal(result, target)


if __name__ == "__main__":
    unittest.main()



# Copyright cf-units contributors
#
# This file is part of cf-units and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

import re

import pytest

import cf_units
from cf_units._udunits2_parser import normalize

testdata = [
    "",
    "1",
    "12",
    "1.2",
    "+1",
    "+1.2",
    "-1",
    "-1.2",
    "-1.2e0",
    "2e6",
    "2e-6",
    "2.e-6",
    ".1e2",
    ".1e2.2",
    "2e",  # <- TODO: Assert this isn't 2e1, but is infact the unit e *2
    "m",
    "meter",
    # Multiplication
    "1 2 3",
    "1 -2 -3",
    "1m",
    "1*m",
    "m·m",
    "1 m",
    "1   m",
    "m -1",
    "m -1.2",
    "m 1",
    "m 1.2",
    "m-+2",
    "m--4",
    "m*1*2",
    "m--2--3",
    # TODO: add some tests with brackets.
    "m(2.3)",
    "m(2.3m)",
    "(1.2)(2.4)",
    "(5m(6s-1))",
    "2*3*4/5m/6*7*8",
    "m/2",
    "m1",
    "m m",
    "m2",
    "m+2",
    "m¹",
    "m²",
    "m³",
    "2⁴",  # NOTE: Udunits can't do m⁴ for some reason. Bug?
    "2⁵",
    "2⁴²",
    "3⁻²",
    "m2 s2",
    "m^2*s^2",
    "1-2",
    "1-2-3",  # nb. looks a bit like a date, but it isn't!
    "m-1",
    "m^2",
    "m^+2",
    "m^-1",
    "m.2",  # This is 2*m
    "m.+2",  # 2*m
    "m.2.4",  # This is 2.4 * m
    "m0.2",  # But this is 2 m^0
    "m2.5",  # And this is 5m^2
    "m2.3.4",  # 0.4 * m^2
    "m--1",
    # Division
    "m per 2",
    "m per s",
    "m / 2",
    # Shift
    "m@10",
    "m @10",
    "m @ 10",
    "m@ 10",
    "m from2",
    "m from2e-1",
    "(m @ 10) (s @ 10)",
    # Date shift
    "s from 1990",
    "minutes since 1990",
    "hour@1990",
    "hours from 1990-1",
    "hours from 1990-1-1",
    "hours from 1990-1-1 0",
    "hours from 1990-1-1 0:1:1",
    "hours from 1990-1-1 0:0:1 +2",
    "s since 1990-1-2+5:2:2",
    "s since 1990-1-2+5:2",
    "s since 1990-1-2 5 6:0",  # Undocumented packed_clock format?
    "s since 19900102T5",  # Packed format (undocumented?)
    "s since 19900101T190030 +2",
    "s since 199022T1",  # UGLY! (bug?).
    "s since 1990 +2:0:2.9",
    "s since 1990-2T1",
    "hours from 1990-1-1 -19:4:2",
    "hours from 1990-1-1 3+1",
    "seconds from 1990-1-1 0:0:0 +2550",
    "s since 1990-1-2+5:2:2",
    "hours from 1990-1-1 0:1:60",
    "hours from 1990-1-1 0:1:62",
    "(hours since 1900) (s since 1980)",  # Really fruity behaviour.
    # Unicode / constants
    "π",
    "e",
    "°C",
]

invalid = [
    "1 * m",
    "m--m",
    "-m",
    ".1e2.",
    "m+-1",
    "--1",
    "+-1",
    "--3.1",
    "$",
    "£",
    "hours from 1990-0-0 0:0:0",
    "hours since 1900-1 10:12 10:0 1",
    "s since 1990:01:02T1900 +1",
]


@pytest.mark.parametrize("_, unit_str", enumerate(testdata))
def test_normed_units_equivalent(_, unit_str):
    # nb: The "_" argument makes it easier to see which test was being run.

    # Get the udunits symbolic form for the raw unit.
    raw_symbol = cf_units.Unit(unit_str).symbol

    # Now get the parsed form of the unit, and then convert that to
    # symbolic form. The two should match.
    unit_expr = normalize(unit_str)
    parsed_expr_symbol = cf_units.Unit(unit_expr).symbol

    # Whilst the symbolic form from udunits is ugly, it *is* acurate,
    # so check that the two represent the same unit.
    assert raw_symbol == parsed_expr_symbol


udunits_bugs = ["2¹²³⁴⁵⁶⁷⁸⁹⁰", "m⁻²"]


@pytest.mark.parametrize("_, unit_str", enumerate(invalid))
def test_invalid_units(_, unit_str):
    # Confirm that invalid udunits-2 units are also invalid in our grammar.

    try:
        cf_units.Unit(unit_str)
        cf_valid = True
    except ValueError:
        cf_valid = False

    # Double check that udunits2 can't parse this.
    assert (
        cf_valid is False
    ), "Unit {!r} is unexpectedly valid in UDUNITS2".format(unit_str)

    try:
        normalize(unit_str)
        can_parse = True
    except SyntaxError:
        can_parse = False

    # Now confirm that we couldn't parse this either.
    msg = "Parser unexpectedly able to deal with {}".format(unit_str)
    assert can_parse is False, msg


def multi_enumerate(items):
    # Like enumerate, but flattens out the resulting index and items.
    return [[i, *item] for i, item in enumerate(items)]


not_udunits = [
    ["foo", "foo"],
    ["mfrom1", "mfrom^1"],
    ["m⁴", "m^4"],  # udunits bug.
    ["2¹²³⁴⁵⁶⁷⁸⁹⁰", "2^1234567890"],
    # Unicode (subset of the subset).
    ["À"] * 2,
    ["Á"] * 2,
    ["Ö"] * 2,
    ["Ø"] * 2,
    ["ö"] * 2,
    ["ø"] * 2,
    ["ÿ"] * 2,
    ["µ"] * 2,
    ["µ°F·Ω⁻¹", "µ°F·Ω^-1"],
]


@pytest.mark.parametrize("_, unit_str, expected", multi_enumerate(not_udunits))
def test_invalid_in_udunits_but_still_parses(_, unit_str, expected):
    # Some units read fine in our grammar, but not in UDUNITS.

    try:
        cf_units.Unit(unit_str)
        cf_valid = True
    except ValueError:
        cf_valid = False

    # Double check that udunits2 can't parse this.
    assert cf_valid is False

    unit_expr = normalize(unit_str)
    assert unit_expr == expected


known_issues = [
    # Disabled due to crazy results from UDUNITS.
    ["s since +1990 +2:0:2.9", SyntaxError],
    ["s since -1990 +2:0:2.9", SyntaxError],
    # The following are not yet implemented.
    ["hours since 2001-12-31 23:59:59.999UTC", SyntaxError],
    ["hours since 2001-12-31 23:59:59.999 Z", SyntaxError],
    ["hours since 2001-12-31 23:59:59.999 GMT", SyntaxError],
    ["0.1 lg(re 1 mW)", SyntaxError],
]


@pytest.mark.parametrize(
    "_, unit_str, expected", multi_enumerate(known_issues)
)
def test_known_issues(_, unit_str, expected):
    # Unfortunately the grammar is not perfect.
    # These are the cases that don't work yet but which do work with udunits.

    # Make sure udunits can read it.
    cf_units.Unit(unit_str).symbol

    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(SyntaxError):
            unit_expr = normalize(unit_str)
    else:
        unit_expr = normalize(unit_str)
        assert unit_expr != expected


def test_syntax_parse_error_quality():
    # Check that the syntax error is giving us good context.

    msg = re.escape(r"no viable alternative at input 'm^m' (inline, line 1)")
    with pytest.raises(SyntaxError, match=msg) as err:
        normalize("m^m 2s")
    # The problem is with the m after "^", so make sure the exception is
    # pointing at it (including the leading speechmark).
    assert err.value.offset == 4


def test_unknown_symbol_error():
    msg = re.escape(r"mismatched input '×' expecting <EOF>")
    with pytest.raises(SyntaxError, match=msg) as err:
        # The × character is explicitly excluded in the UDUNITS2
        # implementation. It would make some sense to support it in the
        # future though.
        normalize("Thing×Another")
    # The 7th character (including the speechmark) is the problem, check that
    # the exception points at the right location.
    # correct location...
    #  File "inline", line 1
    #  'Thing×Another'
    #        ^
    assert err.value.offset == 7


not_allowed = [
    "hours from 1990-1-1 -20:4:18 +2",
    "m++2",
    "m s^(-1)",
    "m per /s",
]


@pytest.mark.parametrize("_, unit_str", enumerate(not_allowed))
def test_invalid_syntax_units(_, unit_str):
    # Check that units that aren't allowed with UDUNITS-2 are also not
    # allowed with our grammar.

    with pytest.raises(ValueError):
        cf_units.Unit(unit_str).symbol

    with pytest.raises(SyntaxError):
        normalize(unit_str)



# Copyright cf-units contributors
#
# This file is part of cf-units and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.


import configparser
import os.path
import sys


# Returns simple string options.
def get_option(section, option, default=None):
    """
    Returns the option value for the given section, or the default value
    if the section/option is not present.

    """
    value = default
    if config.has_option(section, option):
        value = config.get(section, option)
    return value


# Figure out the full path to the "cf_units" package.
ROOT_PATH = os.path.abspath(os.path.dirname(__file__))

# The full path to the configuration directory of the active cf_units instance.
CONFIG_PATH = os.path.join(ROOT_PATH, "etc")

# Load the optional "site.cfg" file if it exists.
if sys.version_info >= (3, 2):
    config = configparser.ConfigParser()
else:
    config = configparser.SafeConfigParser()
config.read([os.path.join(CONFIG_PATH, "site.cfg")])



# Copyright cf-units contributors
#
# This file is part of cf-units and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Units of measure.

Provision of a wrapper class to support Unidata/UCAR UDUNITS-2, and the
cftime calendar functionality.

See also: `UDUNITS-2
<http://www.unidata.ucar.edu/software/udunits>`_.

"""

import copy
import os.path
import sys
from contextlib import contextmanager

import cftime
import numpy as np

from cf_units import _udunits2 as _ud
from cf_units._udunits2 import (
    UT_ASCII,
    UT_DEFINITION,
    UT_ISO_8859_1,
    UT_LATIN1,
    UT_NAMES,
    UT_UTF8,
)

from . import config
from ._version import version as __version__  # noqa: F401
from .util import _OrderedHashable, approx_equal

__all__ = [
    "CALENDAR_STANDARD",
    "CALENDAR_GREGORIAN",
    "CALENDAR_PROLEPTIC_GREGORIAN",
    "CALENDAR_NO_LEAP",
    "CALENDAR_JULIAN",
    "CALENDAR_ALL_LEAP",
    "CALENDAR_365_DAY",
    "CALENDAR_366_DAY",
    "CALENDAR_360_DAY",
    "CALENDARS",
    "CALENDAR_ALIASES",
    "UT_NAMES",
    "UT_DEFINITION",
    "UT_ASCII",
    "FLOAT32",
    "FLOAT64",
    "is_time",
    "is_vertical",
    "Unit",
    "date2num",
    "decode_time",
    "encode_clock",
    "encode_date",
    "encode_time",
    "num2date",
    "num2pydate",
    "suppress_errors",
]


########################################################################
#
# module level constants
#
########################################################################

#
# default constants
#
EPOCH = "1970-01-01 00:00:00"
_UNKNOWN_UNIT_STRING = "unknown"
_UNKNOWN_UNIT_SYMBOL = "?"
_UNKNOWN_UNIT = [_UNKNOWN_UNIT_STRING, _UNKNOWN_UNIT_SYMBOL, "???", ""]
_NO_UNIT_STRING = "no_unit"
_NO_UNIT_SYMBOL = "-"
_NO_UNIT = [_NO_UNIT_STRING, _NO_UNIT_SYMBOL, "no unit", "no-unit", "nounit"]
_UNIT_DIMENSIONLESS = "1"
_OP_SINCE = " since "
_CATEGORY_UNKNOWN, _CATEGORY_NO_UNIT, _CATEGORY_UDUNIT = range(3)


#
# libudunits2 constants
#
UT_FORMATS = [
    UT_ASCII,
    UT_ISO_8859_1,
    UT_LATIN1,
    UT_UTF8,
    UT_NAMES,
    UT_DEFINITION,
]

#
# cftime constants
#
CALENDAR_STANDARD = "standard"
CALENDAR_GREGORIAN = "gregorian"
CALENDAR_PROLEPTIC_GREGORIAN = "proleptic_gregorian"
CALENDAR_NO_LEAP = "noleap"
CALENDAR_JULIAN = "julian"
CALENDAR_ALL_LEAP = "all_leap"
CALENDAR_365_DAY = "365_day"
CALENDAR_366_DAY = "366_day"
CALENDAR_360_DAY = "360_day"

#: The calendars recognised by cf_units.
#: These are accessible as strings, or as constants in the form
#: ``cf_units.CALENDAR_{ calendar_name.upper() }``. For example,
#: ``cf_units.CALENDAR_NO_LEAP`` and ``cf_units.CALENDAR_366_DAY``.
CALENDARS = [
    CALENDAR_STANDARD,
    CALENDAR_GREGORIAN,
    CALENDAR_PROLEPTIC_GREGORIAN,
    CALENDAR_NO_LEAP,
    CALENDAR_JULIAN,
    CALENDAR_ALL_LEAP,
    CALENDAR_365_DAY,
    CALENDAR_366_DAY,
    CALENDAR_360_DAY,
]

#: Where calendars have multiple names, we map the alias to the
#: definitive form.
CALENDAR_ALIASES = {
    CALENDAR_STANDARD: CALENDAR_GREGORIAN,
    CALENDAR_NO_LEAP: CALENDAR_365_DAY,
    CALENDAR_ALL_LEAP: CALENDAR_366_DAY,
}


#
# floating point types
#
FLOAT32 = np.float32
FLOAT64 = np.float64

########################################################################
#
# module level statements
#
########################################################################

# Convenience dictionary for the Unit convert method.
_cv_convert_scalar = {FLOAT32: _ud.convert_float, FLOAT64: _ud.convert_double}
_cv_convert_array = {FLOAT32: _ud.convert_floats, FLOAT64: _ud.convert_doubles}

# Map of ut_encodings to encoding strings
_encoding_lookup = {
    UT_ASCII: "ascii",
    UT_ISO_8859_1: "iso_8859_1",
    UT_LATIN1: "latin1",
    UT_UTF8: "utf-8",
}


@contextmanager
def suppress_errors():
    """
    Suppresses all error messages from UDUNITS-2.

    """
    _default_handler = _ud.set_error_message_handler(_ud.ignore)
    try:
        yield
    finally:
        _ud.set_error_message_handler(_default_handler)


#
# load the UDUNITS-2 xml-formatted unit-database
#:
# Ignore standard noisy UDUNITS-2 start-up.
with suppress_errors():
    # Load the unit-database from the default location (modified via
    # the UDUNITS2_XML_PATH environment variable) and if that fails look
    # relative to sys.prefix to support environments such as conda.
    try:
        _ud_system = _ud.read_xml()
    except _ud.UdunitsError:
        _alt_xml_path = config.get_option(
            "System",
            "udunits2_xml_path",
            default=os.path.join(
                sys.prefix, "share", "udunits", "udunits2.xml"
            ),
        )
        try:
            _ud_system = _ud.read_xml(_alt_xml_path.encode())
        except _ud.UdunitsError as e:
            error_msg = ': "%s"' % e.error_msg() if e.errnum else ""
            raise OSError(
                "[%s] Failed to open UDUNITS-2 XML unit database%s"
                % (e.status_msg(), error_msg)
            )


########################################################################
#
# module level function definitions
#
########################################################################


def encode_time(year, month, day, hour, minute, second):
    """
    Return date/clock time encoded as a double precision value.

    Encoding performed using UDUNITS-2 hybrid Gregorian/Julian calendar.
    Dates on or after 1582-10-15 are assumed to be Gregorian dates;
    dates before that are assumed to be Julian dates. In particular, the
    year 1 BCE is immediately followed by the year 1 CE.

    Args:

    * year (int):
        Year value to be encoded.
    * month (int):
        Month value to be encoded.
    * day (int):
        Day value to be encoded.
    * hour (int):
        Hour value to be encoded.
    * minute (int):
        Minute value to be encoded.
    * second (int):
        Second value to be encoded.

    Returns:
        float.

    For example:

        >>> import cf_units
        >>> cf_units.encode_time(1970, 1, 1, 0, 0, 0)
        -978307200.0

    """

    return _ud.encode_time(year, month, day, hour, minute, second)


def encode_date(year, month, day):
    """
    Return date encoded as a double precision value.

    Encoding performed using UDUNITS-2 hybrid Gergorian/Julian calendar.
    Dates on or after 1582-10-15 are assumed to be Gregorian dates;
    dates before that are assumed to be Julian dates. In particular, the
    year 1 BCE is immediately followed by the year 1 CE.

    Args:

    * year (int):
        Year value to be encoded.
    * month (int):
        Month value to be encoded.
    * day (int):
        Day value to be encoded.

    Returns:
        float.

    For example:

        >>> import cf_units
        >>> cf_units.encode_date(1970, 1, 1)
        -978307200.0

    """

    return _ud.encode_date(year, month, day)


def encode_clock(hour, minute, second):
    """
    Return clock time encoded as a double precision value.

    Args:

    * hour (int):
        Hour value to be encoded.
    * minute (int):
        Minute value to be encoded.
    * second (int):
        Second value to be encoded.

    Returns:
        float.

    For example:

        >>> import cf_units
        >>> cf_units.encode_clock(0, 0, 0)
        0.0

    """

    return _ud.encode_clock(hour, minute, second)


def decode_time(time):
    """
    Decode a double precision date/clock time value into its component
    parts and return as tuple.

    Decode time into it's year, month, day, hour, minute, second, and
    resolution component parts. Where resolution is the uncertainty of
    the time in seconds.

    Args:

    * time (float): Date/clock time encoded as a double precision value.

    Returns:
        tuple of (year, month, day, hour, minute, second, resolution).

    For example:

        >>> import cf_units
        >>> cf_units.decode_time(cf_units.encode_time(1970, 1, 1, 0, 0, 0))
        (1970, 1, 1, 0, 0, 0.0, 1.086139178596568e-07)

    """
    return _ud.decode_time(time)


def date2num(date, unit, calendar):
    """
    Return numeric time value (resolution of 1 second) encoding of
    datetime object.

    The units of the numeric time values are described by the unit and
    calendar arguments. The datetime objects must be in UTC with no
    time-zone offset. If there is a time-zone offset in unit, it will be
    applied to the returned numeric values.

    Args:

    * date (datetime):
        A datetime object or a sequence of datetime objects.
        The datetime objects should not include a time-zone offset.
    * unit (string):
        A string of the form '<time-unit> since <time-origin>' describing
        the time units. The <time-unit> can be days, hours, minutes or seconds.
        The <time-origin> is a date/time reference point. A valid choice
        would be unit='hours since 1800-01-01 00:00:00 -6:00'.
    * calendar (string):
        Name of the calendar, see cf_units.CALENDARS.

    Returns:
        float, or numpy.ndarray of float.

    For example:

        >>> import cf_units
        >>> import datetime
        >>> dt1 = datetime.datetime(1970, 1, 1, 6, 30, 0)
        >>> dt2 = datetime.datetime(1970, 1, 1, 7, 30, 0)
        >>> cf_units.date2num(dt1, 'hours since 1970-01-01 00:00:00',
        ...               cf_units.CALENDAR_STANDARD)
        6.5
        >>> cf_units.date2num([dt1, dt2], 'hours since 1970-01-01 00:00:00',
        ...               cf_units.CALENDAR_STANDARD)
        array([6.5, 7.5])

    """

    #
    # ensure to strip out any 'UTC' postfix which is generated by
    # UDUNITS-2 formatted output and causes the cftime parser
    # to choke
    #
    unit_string = unit.rstrip(" UTC")
    if unit_string.endswith(" since epoch"):
        unit_string = unit_string.replace("epoch", EPOCH)
    unit_inst = Unit(unit_string, calendar=calendar)
    return unit_inst.date2num(date)


def _discard_microsecond(date):
    """
    Return a date with the microsecond component discarded.

    Works for scalars, sequences and numpy arrays. Returns a scalar
    if input is a scalar, else returns a numpy array.

    Args:

    * date (datetime.datetime or cftime.datetime):
        Date value/s

    Returns:
        datetime, or numpy.ndarray of datetime object.

    """
    dates = np.asanyarray(date)
    shape = dates.shape
    dates = dates.ravel()

    # using the "and" pattern to support masked arrays of datetimes
    dates = np.array([dt and dt.replace(microsecond=0) for dt in dates])
    result = dates[0] if shape == () else dates.reshape(shape)

    return result


def num2date(
    time_value,
    unit,
    calendar,
    only_use_cftime_datetimes=True,
    only_use_python_datetimes=False,
):
    """
    Return datetime encoding of numeric time value (resolution of 1 second).

    The units of the numeric time value are described by the unit and
    calendar arguments. The returned datetime object represent UTC with
    no time-zone offset, even if the specified unit contain a time-zone
    offset.

    By default, the datetime instances returned are cftime.datetime objects,
    regardless of calendar.  If the only_use_cftime_datetimes keyword is set to
    False, they are datetime.datetime objects if the date falls in the
    Gregorian calendar (i.e. calendar is 'proleptic_gregorian', 'standard' or
    'gregorian' and the date is after 1582-10-15). The datetime instances
    do not contain a time-zone offset, even if the specified unit
    contains one.

    Works for scalars, sequences and numpy arrays. Returns a scalar
    if input is a scalar, else returns a numpy array.

    Args:

    * time_value (float):
        Numeric time value/s. Maximum resolution is 1 second.
    * unit (sting):
        A string of the form '<time-unit> since <time-origin>'
        describing the time units. The <time-unit> can be days, hours,
        minutes or seconds. The <time-origin> is the date/time reference
        point. A valid choice would be
        unit='hours since 1800-01-01 00:00:00 -6:00'.
    * calendar (string):
        Name of the calendar, see cf_units.CALENDARS.

    Kwargs:

    * only_use_cftime_datetimes (bool):
        If True, will always return cftime datetime objects, regardless of
        calendar.  If False, returns datetime.datetime instances where
        possible.  Defaults to True.

    * only_use_python_datetimes (bool):
        If True, will always return datetime.datetime instances where
        possible, and raise an exception if not.  Ignored if
        only_use_cftime_datetimes is True.  Defaults to False.

    Returns:
        datetime, or numpy.ndarray of datetime object.

    For example:

        >>> import cf_units
        >>> import datetime
        >>> print(cf_units.num2date(6, 'hours since 1970-01-01 00:00:00',
        ...                         cf_units.CALENDAR_STANDARD))
        1970-01-01 06:00:00
        >>> dts = cf_units.num2date([6, 7], 'hours since 1970-01-01 00:00:00',
        ...                         cf_units.CALENDAR_STANDARD)
        >>> [str(dt) for dt in dts]
        ['1970-01-01 06:00:00', '1970-01-01 07:00:00']

    """

    #
    # ensure to strip out any 'UTC' postfix which is generated by
    # UDUNITS-2 formatted output and causes the cftime parser
    # to choke
    #
    unit_string = unit.rstrip(" UTC")
    if unit_string.endswith(" since epoch"):
        unit_string = unit_string.replace("epoch", EPOCH)
    unit_inst = Unit(unit_string, calendar=calendar)
    return unit_inst.num2date(
        time_value,
        only_use_cftime_datetimes=only_use_cftime_datetimes,
        only_use_python_datetimes=only_use_python_datetimes,
    )


def num2pydate(time_value, unit, calendar):
    """
    Convert time value(s) to python datetime.datetime objects, or raise an
    exception if this is not possible.  Same as::

        num2date(time_value, unit, calendar,
                 only_use_cftime_datetimes=False,
                 only_use_python_datetimes=True)

    """
    return num2date(
        time_value,
        unit,
        calendar,
        only_use_cftime_datetimes=False,
        only_use_python_datetimes=True,
    )


def _num2date_to_nearest_second(
    time_value,
    unit,
    only_use_cftime_datetimes=True,
    only_use_python_datetimes=False,
):
    """
    Return datetime encoding of numeric time value with respect to the given
    time reference units, with a resolution of 1 second.

    * time_value (float):
        Numeric time value/s.
    * unit (Unit):
        cf_units.Unit object with which to perform the conversion/s.

    * only_use_cftime_datetimes (bool):
        If True, will always return cftime datetime objects, regardless of
        calendar.  If False, returns datetime.datetime instances where
        possible.  Defaults to True.

    * only_use_python_datetimes (bool):
        If True, will always return datetime.datetime instances where
        possible, and raise an exception if not.  Ignored if
        only_use_cftime_datetimes is True.  Defaults to False.

    Returns:
        datetime, or numpy.ndarray of datetime object.

    """
    time_values = np.asanyarray(time_value)
    shape = time_values.shape
    time_values = time_values.ravel()

    # We account for the edge case where the time is in seconds and has a
    # half second: cftime.num2date() may produce a date that would round
    # down.
    #
    # Note that this behaviour is different to the num2date function in version
    # 1.1 and earlier of cftime that didn't have microsecond precision. In
    # those versions, a half-second value would be rounded up or down
    # arbitrarily. It is probably not possible to replicate that behaviour with
    # later versions, if one wished to do so for the sake of consistency.
    cftime_unit = unit.cftime_unit
    time_units = cftime_unit.split(" ")[0]
    has_half_seconds = np.logical_and(
        time_units == "seconds", time_values % 1.0 == 0.5
    )
    num2date_kwargs = dict(
        units=cftime_unit,
        calendar=unit.calendar,
        only_use_cftime_datetimes=only_use_cftime_datetimes,
        only_use_python_datetimes=only_use_python_datetimes,
    )
    dates = cftime.num2date(time_values, **num2date_kwargs)
    try:
        # We can assume all or none of the dates have a microsecond attribute
        microseconds = np.array([d.microsecond if d else 0 for d in dates])
    except AttributeError:
        microseconds = 0
    round_mask = np.logical_or(has_half_seconds, microseconds != 0)
    ceil_mask = np.logical_or(has_half_seconds, microseconds >= 500000)
    if time_values[ceil_mask].size > 0:
        useconds = Unit("second")
        second_frac = useconds.convert(0.75, time_units)
        dates[ceil_mask] = cftime.num2date(
            time_values[ceil_mask] + second_frac, **num2date_kwargs
        )
    dates[round_mask] = _discard_microsecond(dates[round_mask])
    result = dates[0] if shape == () else dates.reshape(shape)
    return result


_CACHE = {}


def as_unit(unit):
    """
    Returns a Unit corresponding to the given unit.

    .. note::

        If the given unit is already a Unit it will be returned unchanged.

    """
    if isinstance(unit, Unit):
        result = unit
    else:
        result = None
        use_cache = isinstance(unit, (str,)) or unit is None
        if use_cache:
            result = _CACHE.get(unit)
        if result is None:
            # Typically unit is a string, however we cater for other types of
            # 'unit' (e.g. iris.unit.Unit).
            result = Unit(unit, calendar=getattr(unit, "calendar", None))
            if use_cache:
                _CACHE[unit] = result
    return result


def is_time(unit):
    """
    Determine whether the unit is a related SI Unit of time.

    Args:

    * unit (string/Unit): Unit to be compared.

    Returns:
        Boolean.

    For example:

        >>> import cf_units
        >>> cf_units.is_time('hours')
        True
        >>> cf_units.is_time('meters')
        False

    """
    return as_unit(unit).is_time()


def is_vertical(unit):
    """
    Determine whether the unit is a related SI Unit of pressure or distance.

    Args:

    * unit (string/Unit): Unit to be compared.

    Returns:
        Boolean.

    For example:

        >>> import cf_units
        >>> cf_units.is_vertical('millibar')
        True
        >>> cf_units.is_vertical('km')
        True

    """
    return as_unit(unit).is_vertical()


def _ud_value_error(ud_err, message):
    """
    Return a ValueError that has extra context from a _udunits2.UdunitsError.

    """
    # NOTE: We aren't raising here, just giving the caller a well formatted
    # exception that they can raise themselves.

    ud_msg = ud_err.error_msg()
    if ud_msg:
        message = "{}: {}".format(message, ud_msg)

    message = "[{status}] {message}".format(
        status=ud_err.status_msg(), message=message
    )

    return ValueError(message)


class Unit(_OrderedHashable):
    """
    A class to represent S.I. units and support common operations to
    manipulate such units in a consistent manner as per UDUNITS-2.

    These operations include scaling the unit, offsetting the unit by a
    constant or time, inverting the unit, raising the unit by a power,
    taking a root of the unit, taking a log of the unit, multiplying the
    unit by a constant or another unit, dividing the unit by a constant
    or another unit, comparing units, copying units and converting unit
    data to single precision or double precision floating point numbers.

    This class also supports time and calendar defintion and manipulation.

    """

    def _init_from_tuple(self, values):
        # Implements the required interface for an _OrderedHashable.
        # This will also ensure a Unit._init(*Unit.names) method exists.
        for name, value in zip(self._names, values):
            object.__setattr__(self, name, value)

    # Provide hash semantics

    def _identity(self):
        return (self.name, self.calendar)

    def __hash__(self):
        return hash(self._identity())

    # Provide default ordering semantics

    def __lt__(self, other):
        return self._identity() < other._identity()

    # Prevent attribute updates

    def __setattr__(self, name, value):
        raise AttributeError(
            "Instances of %s are immutable" % type(self).__name__
        )

    def __delattr__(self, name):
        raise AttributeError(
            "Instances of %s are immutable" % type(self).__name__
        )

    # Declare the attribute names relevant to the ordered and hashable
    #  behaviour.
    _names = ("category", "ut_unit", "calendar", "origin")

    category = None
    "Is this an unknown unit, a no-unit, or a UDUNITS-2 unit."

    ut_unit = None
    "Reference to the quantity defining the UDUNITS-2 unit."

    calendar = None
    "Represents the unit calendar name, see cf_units.CALENDARS"

    origin = None
    "The original string used to create this unit."

    __slots__ = ()

    def __init__(self, unit, calendar=None):
        """
        Create a wrapper instance for UDUNITS-2.

        An optional calendar may be provided for a unit which defines a
        time reference of the form '<time-unit> since <time-origin>'
        i.e. unit='days since 1970-01-01 00:00:00'. For a unit that is a
        time reference, the default calendar is 'standard'.

        Accepted calendars are as follows,

        * 'standard' or 'gregorian' - Mixed Gregorian/Julian calendar as
          defined by udunits.
        * 'proleptic_gregorian' - A Gregorian calendar extended to dates
          before 1582-10-15. A year is a leap year if either,

            1. It is divisible by 4 but not by 100, or
            2. It is divisible by 400.

        * 'noleap' or '365_day' - A Gregorian calendar without leap
          years i.e. all years are 365 days long.
        * 'all_leap' or '366_day' - A Gregorian calendar with every year
          being a leap year i.e. all years are 366 days long.
        * '360_day' - All years are 360 days divided into 30 day months.
        * 'julian' - Proleptic Julian calendar, extended to dates after
          1582-10-5. A year is a leap year if it is divisible by 4.

        Args:

        * unit:
            Specify the unit as defined by UDUNITS-2.
        * calendar (string):
            Describes the calendar used in time calculations. The
            default is 'standard' or 'gregorian' for a time reference
            unit.

        Returns:

            Unit object.

        Units should be set to "no_unit" for values which are strings.
        Units can also be set to "unknown" (or None).
        For example:

            >>> from cf_units import Unit
            >>> volts = Unit('volts')
            >>> no_unit = Unit('no_unit')
            >>> unknown = Unit('unknown')
            >>> unknown = Unit(None)

        """
        ut_unit = _ud.NULL_UNIT
        calendar_ = None

        encoding = UT_UTF8

        if unit is None:
            unit = ""

        unit = str(unit).strip()

        if unit.lower().endswith(" utc"):
            unit = unit[: unit.lower().rfind(" utc")]

        if unit.endswith(" since epoch"):
            unit = unit.replace("epoch", EPOCH)

        if "#" in unit:
            unit = unit.replace("#", "1")

        if unit.lower() in _UNKNOWN_UNIT:
            # TODO - removing the option of an unknown unit. Currently
            # the auto generated MOSIG rules are missing units on a
            # number of phenomena which would lead to errors.
            # Will be addressed by work on metadata translation.
            category = _CATEGORY_UNKNOWN
            unit = _UNKNOWN_UNIT_STRING
        elif unit.lower() in _NO_UNIT:
            category = _CATEGORY_NO_UNIT
            unit = _NO_UNIT_STRING
        else:
            category = _CATEGORY_UDUNIT
            str_unit = unit
            try:
                ut_unit = _ud.parse(_ud_system, unit.encode("utf8"), encoding)
            except _ud.UdunitsError as exception:
                value_error = _ud_value_error(
                    exception, 'Failed to parse unit "{}"'.format(str_unit)
                )
                raise value_error from None
            if _OP_SINCE in unit.lower():
                if calendar is None:
                    calendar_ = CALENDAR_GREGORIAN
                elif isinstance(calendar, (str,)):
                    calendar_ = calendar.lower()
                    if calendar_ in CALENDAR_ALIASES:
                        calendar_ = CALENDAR_ALIASES[calendar_]
                    if calendar_ not in CALENDARS:
                        msg = "{!r} is an unsupported calendar."
                        raise ValueError(msg.format(calendar))
                else:
                    msg = "Expected string-like calendar argument, got {!r}."
                    raise TypeError(msg.format(type(calendar)))

        # Call the OrderedHashable's init.
        self._init(
            category,
            ut_unit,
            calendar_,
            unit,
        )

    @classmethod
    def _new_from_existing_ut(
        cls, category, ut_unit, calendar=None, origin=None
    ):
        # Short-circuit __init__ if we know what we are doing and already
        # have a UT handle.
        unit = cls.__new__(cls)
        unit._init(category, ut_unit, calendar, origin)
        return unit

    # NOTE:
    # "__getstate__" and "__setstate__" functions are defined here to
    # provide a custom interface for Pickle
    #  : Pickle "normal" behaviour is just to save/reinstate the object
    #    dictionary
    #  : that won't work here, because the "ut_unit" attribute is an
    #    object handle
    #    - the corresponding udunits object only exists in the original
    #      invocation
    def __getstate__(self):
        # state capture method for Pickle.dump()
        #  - return the instance data needed to reconstruct a Unit value
        return {"unit_text": self.origin, "calendar": self.calendar}

    def __setstate__(self, state):
        # object reconstruction method for Pickle.load()
        # intercept the Pickle.load() operation and call own __init__ again
        #  - this is to ensure a valid ut_unit attribute (as these
        #    handles aren't persistent)
        self.__init__(state["unit_text"], calendar=state["calendar"])

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    def is_time(self):
        """
        Determine whether this unit is a related SI Unit of time.

        Returns:
            Boolean.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('hours')
            >>> u.is_time()
            True
            >>> v = cf_units.Unit('meter')
            >>> v.is_time()
            False

        """
        if self.is_unknown() or self.is_no_unit():
            result = False
        else:
            day = _ud.get_unit_by_name(_ud_system, b"day")
            result = _ud.are_convertible(self.ut_unit, day)
        return result

    def is_vertical(self):
        """
        Determine whether the unit is a related SI Unit of pressure or
        distance.

        Returns:
            Boolean.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('millibar')
            >>> u.is_vertical()
            True
            >>> v = cf_units.Unit('km')
            >>> v.is_vertical()
            True

        """
        if self.is_unknown() or self.is_no_unit():
            result = False
        else:
            bar = _ud.get_unit_by_name(_ud_system, b"bar")
            result = _ud.are_convertible(self.ut_unit, bar)
            if not result:
                meter = _ud.get_unit_by_name(_ud_system, b"meter")
                result = _ud.are_convertible(self.ut_unit, meter)
        return result

    def is_udunits(self):
        """Return whether the unit is a vaild unit of UDUNITS."""
        return self.ut_unit is not _ud.NULL_UNIT

    def is_time_reference(self):
        """
        Return whether the unit is a time reference unit of the form
        '<time-unit> since <time-origin>'
        i.e. unit='days since 1970-01-01 00:00:00'

        Returns:
            Boolean.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('days since epoch')
            >>> u.is_time_reference()
            True

        """
        return self.calendar is not None

    def is_long_time_interval(self):
        """
        Defines whether this unit describes a time unit with a long time
        interval ("months" or "years"). These long time intervals *are*
        supported by `UDUNITS2` but are not supported by `cftime`. This
        discrepancy means we cannot run self.num2date() on a time unit with
        a long time interval.

        Returns:
            Boolean.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('days since epoch')
            >>> u.is_long_time_interval()
            False
            >>> u = cf_units.Unit('years since epoch')
            >>> u.is_long_time_interval()
            True

        """
        result = False
        long_time_intervals = ["year", "month"]
        if self.is_time_reference():
            result = any(
                interval in self.origin for interval in long_time_intervals
            )
        return result

    def title(self, value):
        """
        Return the unit value as a title string.

        Args:

        * value (float): Unit value to be incorporated into title string.

        Returns:
            string.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('hours since epoch',
            ...                   calendar=cf_units.CALENDAR_STANDARD)
            >>> u.title(10)
            '1970-01-01 10:00:00'

        """
        if self.is_time_reference():
            dt = self.num2date(value)
            result = dt.strftime("%Y-%m-%d %H:%M:%S")
        else:
            result = "%s %s" % (str(value), self)
        return result

    @property
    def modulus(self):
        """
        *(read-only)* Return the modulus value of the unit.

        Convenience method that returns the unit modulus value as follows,
            * 'radians' - pi*2
            * 'degrees' - 360.0
            * Otherwise None.

        Returns:
            float.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('degrees')
            >>> u.modulus
            360.0

        """

        if self == "radians":
            result = np.pi * 2
        elif self == "degrees":
            result = 360.0
        else:
            result = None
        return result

    def is_convertible(self, other):
        """
        Return whether two units are convertible.

        Args:

        * other (Unit): Unit to be compared.

        Returns:
            Boolean.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('meters')
            >>> v = cf_units.Unit('kilometers')
            >>> u.is_convertible(v)
            True

        """
        other = as_unit(other)
        if (
            self.is_unknown()
            or self.is_no_unit()
            or other.is_unknown()
            or other.is_no_unit()
        ):
            result = False
        else:
            result = self.calendar == other.calendar and _ud.are_convertible(
                self.ut_unit, other.ut_unit
            )
        return result

    def is_dimensionless(self):
        """
        Return whether the unit is dimensionless.

        Returns:
            Boolean.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('meters')
            >>> u.is_dimensionless()
            False
            >>> u = cf_units.Unit('1')
            >>> u.is_dimensionless()
            True

        """
        return self.category == _CATEGORY_UDUNIT and bool(
            _ud.is_dimensionless(self.ut_unit)
        )

    def is_unknown(self):
        """
        Return whether the unit is defined to be an *unknown* unit.

        Returns:
            Boolean.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('unknown')
            >>> u.is_unknown()
            True
            >>> u = cf_units.Unit('meters')
            >>> u.is_unknown()
            False

        """
        return self.category == _CATEGORY_UNKNOWN or self.ut_unit is None

    def is_no_unit(self):
        """
        Return whether the unit is defined to be a *no_unit* unit.

        Typically, a quantity such as a string, will have no associated
        unit to describe it. Such a class of quantity may be defined
        using the *no_unit* unit.

        Returns:
            Boolean.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('no unit')
            >>> u.is_no_unit()
            True
            >>> u = cf_units.Unit('meters')
            >>> u.is_no_unit()
            False

        """
        return self.category == _CATEGORY_NO_UNIT

    def format(self, option=None):
        """
        Return a formatted string representation of the binary unit.

        Args:

        * option (cf_units.UT_FORMATS):
            Set the option of the formatted string representation.
            Valid encoding options may be at most one of the following
            enumerations:
            * Unit.UT_ASCII
            * Unit.UT_ISO_8859_1
            * Unit.UT_LATIN1
            * Unit.UT_UTF8

            Any combination of the following may also be used:
            * Unit.UT_NAMES
            * Unit.UT_DEFINITION

            Multiple options may be combined within a list. The default
            option is cf_units.UT_ASCII.

        Returns:
            string.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('meters')
            >>> u.format()
            'm'
            >>> u.format(cf_units.UT_NAMES)
            'meter'
            >>> u.format(cf_units.UT_DEFINITION)
            'm'

        """
        if self.is_unknown():
            return _UNKNOWN_UNIT_STRING
        elif self.is_no_unit():
            return _NO_UNIT_STRING
        else:
            bitmask = UT_ASCII
            if option is not None:
                if not isinstance(option, list):
                    option = [option]
                for i in option:
                    bitmask |= i
            encoding = bitmask & (
                UT_ASCII | UT_ISO_8859_1 | UT_LATIN1 | UT_UTF8
            )
            encoding_str = _encoding_lookup[encoding]
            result = _ud.format(self.ut_unit, bitmask)

            result = str(result.decode(encoding_str))
            return result

    @property
    def name(self):
        """
        *(read-only)* The full name of the unit.

        Formats the binary unit into a string representation using
        method :func:`cf_units.Unit.format` with keyword argument
        option=cf_units.UT_NAMES.

        Returns:
            string.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('watts')
            >>> u.name
            'watt'

        """
        return self.format(UT_NAMES)

    @property
    def symbol(self):
        """
        *(read-only)* The symbolic representation of the unit.

        Formats the binary unit into a string representation using
        method :func:`cf_units.Unit.format`.

        Returns:
            string.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('watts')
            >>> u.symbol
            'W'

        """
        if self.is_unknown():
            result = _UNKNOWN_UNIT_SYMBOL
        elif self.is_no_unit():
            result = _NO_UNIT_SYMBOL
        else:
            result = self.format()
        return result

    @property
    def definition(self):
        """
        *(read-only)* The symbolic decomposition of the unit.

        Formats the binary unit into a string representation using
        method :func:`cf_units.Unit.format` with keyword argument
        option=cf_units.UT_DEFINITION.

        Returns:
            string.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('watts')
            >>> u.definition
            'm2.kg.s-3'

        """
        if self.is_unknown():
            result = _UNKNOWN_UNIT_SYMBOL
        elif self.is_no_unit():
            result = _NO_UNIT_SYMBOL
        else:
            result = self.format(UT_DEFINITION)
        return result

    def offset_by_time(self, origin):
        """
        Returns the time unit offset with respect to the time origin.

        Args:

        * origin (float): Time origin as returned by the
          :func:`cf_units.encode_time` method.

        Returns:
            None.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('hours')
            >>> u.offset_by_time(cf_units.encode_time(1970, 1, 1, 0, 0, 0))
            Unit('h @ 19700101T000000.0000000 UTC')

        """

        if not isinstance(origin, (float, (int,))):
            raise TypeError(
                "a numeric type for the origin argument is" " required"
            )
        try:
            ut_unit = _ud.offset_by_time(self.ut_unit, origin)
        except _ud.UdunitsError as exception:
            value_error = _ud_value_error(
                exception, "Failed to offset {!r}".format(self)
            )
            raise value_error from None
        calendar = None
        return Unit._new_from_existing_ut(_CATEGORY_UDUNIT, ut_unit, calendar)

    def invert(self):
        """
        Invert the unit i.e. find the reciprocal of the unit, and return
        the Unit result.

        Returns:
            Unit.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('meters')
            >>> u.invert()
            Unit('m-1')

        """
        if self.is_unknown():
            result = self
        elif self.is_no_unit():
            raise ValueError("Cannot invert a 'no-unit'.")
        else:
            ut_unit = _ud.invert(self.ut_unit)
            result = Unit._new_from_existing_ut(
                _CATEGORY_UDUNIT, ut_unit, calendar=None
            )
        return result

    def root(self, root):
        """
        Returns the given root of the unit.

        Args:

        * root (int): Value by which the unit root is taken.

        Returns:
            None.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('meters^2')
            >>> u.root(2)
            Unit('m')

        .. note::

            Taking a fractional root of a unit is not supported.

        """
        if round(root) != root:
            raise TypeError("An integer for the root argument is required")
        if self.is_unknown():
            result = self
        elif self.is_no_unit():
            raise ValueError("Cannot take the root of a 'no-unit'.")
        else:
            # only update the unit if it is not scalar
            if self == Unit("1"):
                result = self
            else:
                try:
                    ut_unit = _ud.root(self.ut_unit, root)
                except _ud.UdunitsError as exception:
                    value_error = _ud_value_error(
                        exception,
                        "Failed to take the root of {!r}".format(self),
                    )
                    raise value_error from None
                calendar = None
                result = Unit._new_from_existing_ut(
                    _CATEGORY_UDUNIT, ut_unit, calendar
                )
        return result

    def log(self, base):
        """
        Returns the logorithmic unit corresponding to the given
        logorithmic base.

        Args:

        * base (int/float): Value of the logorithmic base.

        Returns:
            None.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('meters')
            >>> u.log(2)
            Unit('lb(re 1 m)')

        """
        if self.is_unknown():
            result = self
        elif self.is_no_unit():
            raise ValueError("Cannot take the logarithm of a 'no-unit'.")
        else:
            try:
                ut_unit = _ud.log(base, self.ut_unit)
            except TypeError:
                raise TypeError(
                    "A numeric type for the base argument is " " required"
                )
            except _ud.UdunitsError as exception:
                value_err = _ud_value_error(
                    exception,
                    "Failed to calculate logorithmic base "
                    "of {!r}".format(self),
                )
                raise value_err from None
            calendar = None
            result = Unit._new_from_existing_ut(
                _CATEGORY_UDUNIT, ut_unit, calendar
            )
        return result

    def __str__(self):
        """
        Returns a simple string representation of the unit.

        Returns:
            string.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('miles/hour')
            >>> str(u)
            'miles/hour'

        """
        return self.origin or self.symbol

    def __repr__(self):
        """
        Returns a string representation of the unit object.

        Returns:
            string.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('meters')
            >>> repr(u)
            "Unit('meters')"

        """
        if self.calendar is None:
            result = "{}('{}')".format(self.__class__.__name__, self)
        else:
            result = "{}('{}', calendar='{}')".format(
                self.__class__.__name__, self, self.calendar
            )
        return result

    def _offset_common(self, offset):
        if self.is_unknown():
            result = self
        elif self.is_no_unit():
            raise ValueError("Cannot offset a 'no-unit'.")
        else:
            try:
                ut_unit = _ud.offset(self.ut_unit, offset)
            except TypeError:
                result = NotImplemented
            else:
                result = Unit._new_from_existing_ut(
                    _CATEGORY_UDUNIT, ut_unit, calendar=None
                )
        return result

    def __add__(self, other):
        return self._offset_common(other)

    def __sub__(self, other):
        try:
            other = -other
        except TypeError:
            result = NotImplemented
        else:
            result = self._offset_common(-other)
        return result

    def _op_common(self, other, op_func):
        # Convienience method to create a new unit from an operation between
        # the units 'self' and 'other'.

        op_label = op_func.__name__

        other = as_unit(other)

        if self.is_no_unit() or other.is_no_unit():
            raise ValueError("Cannot %s a 'no-unit'." % op_label)

        if self.is_unknown() or other.is_unknown():
            result = Unit(_UNKNOWN_UNIT_STRING)
        else:
            try:
                ut_unit = op_func(self.ut_unit, other.ut_unit)
            except _ud.UdunitsError as exception:
                value_err = _ud_value_error(
                    exception,
                    "Failed to {} {!r} by {!r}".format(op_label, self, other),
                )
                raise value_err from None
            calendar = None
            result = Unit._new_from_existing_ut(
                _CATEGORY_UDUNIT, ut_unit, calendar
            )
        return result

    def __rmul__(self, other):
        # NB. Because we've subclassed a tuple, we need to define this to
        # prevent the default tuple-repetition behaviour.
        # ie. 2 * ('a', 'b') -> ('a', 'b', 'a', 'b')
        return self * other

    def __mul__(self, other):
        """
        Multiply the self unit by the other scale factor or unit and
        return the Unit result.

        Note that, multiplication involving an 'unknown' unit will always
        result in an 'unknown' unit.

        Args:

        * other (int/float/string/Unit): Multiplication scale
          factor or unit.

        Returns:
            Unit.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('meters')
            >>> v = cf_units.Unit('hertz')
            >>> u*v
            Unit('m.s-1')

        """
        return self._op_common(other, _ud.multiply)

    def __div__(self, other):
        """
        Divide the self unit by the other scale factor or unit and
        return the Unit result.

        Note that, division involving an 'unknown' unit will always
        result in an 'unknown' unit.

        Args:

        * other (int/float/string/Unit): Division scale factor or unit.

        Returns:
            Unit.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('m.s-1')
            >>> v = cf_units.Unit('hertz')
            >>> u / v
            Unit('m')

        """
        return self._op_common(other, _ud.divide)

    def __truediv__(self, other):
        """
        Divide the self unit by the other scale factor or unit and
        return the Unit result.

        Note that, division involving an 'unknown' unit will always
        result in an 'unknown' unit.

        Args:

        * other (int/float/string/Unit): Division scale factor or unit.

        Returns:
            Unit.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('m.s-1')
            >>> v = cf_units.Unit('hertz')
            >>> u / v
            Unit('m')

        """
        return self.__div__(other)

    def __pow__(self, power):
        """
        Raise the unit by the given power and return the Unit result.

        Note that, UDUNITS-2 does not support raising a
        non-dimensionless unit by a fractional power.
        Approximate floating point power behaviour has been implemented
        specifically for cf_units.

        Args:

        * power (int/float): Value by which the unit power is raised.

        Returns:
            Unit.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('meters')
            >>> u**2
            Unit('m2')

        """
        try:
            power = float(power)
        except ValueError:
            raise TypeError(
                "A numeric value is required for the power" " argument."
            )

        if self.is_unknown():
            result = self
        elif self.is_no_unit():
            raise ValueError("Cannot raise the power of a 'no-unit'.")
        elif self == Unit("1"):
            # 1 ** N -> 1
            result = self
        else:
            # UDUNITS-2 does not support floating point raise/root.
            # But if the power is of the form 1/N, where N is an integer
            # (within a certain acceptable accuracy) then we can find the Nth
            # root.
            if not approx_equal(power, 0.0) and abs(power) < 1:
                if not approx_equal(1 / power, round(1 / power)):
                    raise ValueError("Cannot raise a unit by a decimal.")
                root = int(round(1 / power))
                result = self.root(root)
            else:
                # Failing that, check for powers which are (very nearly) simple
                # integer values.
                if not approx_equal(power, round(power)):
                    msg = "Cannot raise a unit by a decimal (got %s)." % power
                    raise ValueError(msg)
                power = int(round(power))

                try:
                    ut_unit = _ud.raise_(self.ut_unit, power)
                except _ud.UdunitsError as exception:
                    value_err = _ud_value_error(
                        exception,
                        "Failed to raise the power of {!r}".format(self),
                    )
                    raise value_err from None
                result = Unit._new_from_existing_ut(_CATEGORY_UDUNIT, ut_unit)
        return result

    def __eq__(self, other):
        """
        Compare the two units for equality and return the boolean result.

        Args:

        * other (string/Unit): Unit to be compared.

        Returns:
            Boolean.

        For example:

            >>> from cf_units import Unit
            >>> Unit('meters') == Unit('millimeters')
            False
            >>> Unit('meters') == 'm'
            True

        """
        try:
            other = as_unit(other)
        except ValueError:
            return NotImplemented

        # Compare category (i.e. unknown, no_unit, etc.).
        if self.category != other.category:
            return False

        # Compare calendar as UDUNITS cannot handle calendars.
        if self.calendar != other.calendar:
            return False

        # Compare UDUNITS.
        res = _ud.compare(self.ut_unit, other.ut_unit)
        return res == 0

    def __ne__(self, other):
        """
        Compare the two units for inequality and return the boolean result.

        Args:

        * other (string/Unit): Unit to be compared.

        Returns:
            Boolean.

        For example:

            >>> from cf_units import Unit
            >>> Unit('meters') != Unit('millimeters')
            True
            >>> Unit('meters') != 'm'
            False

        """
        return not self == other

    def convert(self, value, other, ctype=FLOAT64, inplace=False):
        """
        Converts a single value or NumPy array of values from the current unit
        to the other target unit.

        If the units are not convertible, then no conversion will take place.

        Args:

        * value (int/float/numpy.ndarray):
            Value/s to be converted.
        * other (string/Unit):
            Target unit to convert to.
        * ctype (cf_units.FLOAT32/cf_units.FLOAT64):
            Floating point 32-bit single-precision (cf_units.FLOAT32) or
            64-bit double-precision (cf_units.FLOAT64) used for conversion
            when `value` is not a NumPy array or is a NumPy array composed of
            NumPy integers. The default is 64-bit double-precision conversion.
        * inplace (bool):
            If ``False``, return a deep copy of the value array. If ``True``,
            convert the values in-place. A new array will be created if
            ``value`` is an integer NumPy array.

        Returns:
            float or numpy.ndarray of appropriate float type.

        For example:

            >>> import cf_units
            >>> import numpy as np
            >>> c = cf_units.Unit('deg_c')
            >>> f = cf_units.Unit('deg_f')
            >>> c.convert(0, f)
            31.999999999999886
            >>> c.convert(0, f, cf_units.FLOAT32)
            32.0
            >>> a64 = np.arange(3, dtype=np.float64)
            >>> c.convert(a64, f)
            array([32. , 33.8, 35.6])
            >>> a32 = np.arange(3, dtype=np.float32)
            >>> c.convert(a32, f)
            array([32. , 33.8, 35.6], dtype=float32)

        .. note::

           Conversion between unit calendars is not permitted unless the
           calendars are aliases, see :attr:`cf_units.CALENDAR_ALIASES`.

           >>> from cf_units import Unit
           >>> a = Unit('days since 1850-1-1', calendar='gregorian')
           >>> b = Unit('days since 1851-1-1', calendar='standard')
           >>> a.convert(365.75, b)
           0.75

        """
        other = as_unit(other)

        if self == other:
            return value

        if self.is_convertible(other):
            if inplace:
                result = value
            else:
                result = copy.deepcopy(value)
            # Use cftime for converting reference times that are not using a
            # gregorian calendar as it handles these and udunits does not.
            if (
                self.is_time_reference()
                and self.calendar != CALENDAR_GREGORIAN
            ):
                result_datetimes = cftime.num2date(
                    result, self.cftime_unit, self.calendar
                )
                result = cftime.date2num(
                    result_datetimes, other.cftime_unit, other.calendar
                )
                convert_type = isinstance(
                    value, np.ndarray
                ) and np.issubsctype(value.dtype, np.floating)
                if convert_type:
                    result = result.astype(value.dtype)
            else:
                try:
                    ut_converter = _ud.get_converter(
                        self.ut_unit, other.ut_unit
                    )
                except _ud.UdunitsError as exception:
                    value_err = _ud_value_error(
                        exception,
                        "Failed to convert {!r} to {!r}".format(self, other),
                    )
                    raise value_err from None
                if isinstance(result, np.ndarray):
                    # Can only handle array of np.float32 or np.float64 so
                    # cast array of ints to array of floats of requested
                    # precision.
                    if issubclass(result.dtype.type, np.integer):
                        result = result.astype(ctype)
                    # Convert arrays with explicit endianness to native
                    # endianness: udunits seems to be tripped up by arrays
                    # with endianness other than native.
                    if result.dtype.byteorder != "=":
                        if inplace:
                            raise ValueError(
                                "Unable to convert non-native byte ordered "
                                "array in-place. Consider byte-swapping "
                                "first."
                            )
                        else:
                            result = result.astype(result.dtype.type)
                    # Strict type check of numpy array.
                    if result.dtype.type not in (np.float32, np.float64):
                        raise TypeError(
                            "Expect a numpy array of '%s' or '%s'"
                            % np.float32,
                            np.float64,
                        )
                    ctype = result.dtype.type
                    # Utilise global convenience dictionary
                    # _cv_convert_array to convert our array in 1d form
                    result_tmp = result.ravel(order="A")
                    # Do the actual conversion.
                    _cv_convert_array[ctype](
                        ut_converter, result_tmp, result_tmp
                    )
                    # If result_tmp was a copy, not a view (i.e. not C
                    # contiguous), copy the data back to the original.
                    if not np.shares_memory(result, result_tmp):
                        result_tmp = result_tmp.reshape(
                            result.shape, order="A"
                        )
                        if isinstance(result, np.ma.MaskedArray):
                            result.data[...] = result_tmp
                        else:
                            result[...] = result_tmp
                else:
                    if ctype not in _cv_convert_scalar:
                        raise ValueError(
                            "Invalid target type. Can only "
                            "convert to float or double."
                        )
                    # Utilise global convenience dictionary
                    # _cv_convert_scalar
                    result = _cv_convert_scalar[ctype](ut_converter, result)
            return result
        else:
            raise ValueError(
                "Unable to convert from '%r' to '%r'." % (self, other)
            )

    @property
    def cftime_unit(self):
        """
        Returns a string suitable for passing as a unit to cftime.num2date and
        cftime.date2num.

        """
        if self.calendar is None:
            raise ValueError("Unit has undefined calendar")

        # `cftime` cannot parse long time intervals ("months" or "years").
        if self.is_long_time_interval():
            interval = self.origin.split(" ")[0]
            emsg = (
                'Time units with interval of "months", "years" '
                '(or singular of these) cannot be processed, got "{!s}".'
            )
            raise ValueError(emsg.format(interval))

        #
        # ensure to strip out non-parsable 'UTC' postfix, which
        # is generated by UDUNITS-2 formatted output
        #
        return str(self).rstrip(" UTC")

    def date2num(self, date):
        """
        Returns the numeric time value calculated from the datetime
        object using the current calendar and unit time reference.

        The current unit time reference must be of the form:
        '<time-unit> since <time-origin>'
        i.e. 'hours since 1970-01-01 00:00:00'

        Works for scalars, sequences and numpy arrays. Returns a scalar
        if input is a scalar, else returns a numpy array.

        Args:

        * date (datetime):
            A datetime object or a sequence of datetime objects.
            The datetime objects should not include a time-zone offset.

        Returns:
            float or numpy.ndarray of float.

        For example:

            >>> import cf_units
            >>> import datetime
            >>> u = cf_units.Unit('hours since 1970-01-01 00:00:00',
            ...                   calendar=cf_units.CALENDAR_STANDARD)
            >>> u.date2num(datetime.datetime(1970, 1, 1, 5, 30))
            5.5
            >>> u.date2num([datetime.datetime(1970, 1, 1, 5, 30),
            ...             datetime.datetime(1970, 1, 1, 6, 30)])
            array([5.5, 6.5])

        """

        date = _discard_microsecond(date)
        return cftime.date2num(date, self.cftime_unit, self.calendar)

    def num2date(
        self,
        time_value,
        only_use_cftime_datetimes=True,
        only_use_python_datetimes=False,
    ):
        """
        Returns a datetime-like object calculated from the numeric time
        value using the current calendar and the unit time reference.

        The current unit time reference must be of the form:
        '<time-unit> since <time-origin>'
        i.e. 'hours since 1970-01-01 00:00:00'

        By default, the datetime instances returned are cftime.datetime
        objects, regardless of calendar.  If the only_use_cftime_datetimes
        keyword is set to False, they are datetime.datetime objects if the date
        falls in the Gregorian calendar (i.e. calendar is
        'proleptic_gregorian', 'standard' or gregorian' and the date is after
        1582-10-15). The datetime instances do not contain a time-zone offset,
        even if the specified unit contains one.

        Works for scalars, sequences and numpy arrays. Returns a scalar
        if input is a scalar, else returns a numpy array.

        Args:

        * time_value (float):
            Numeric time value/s. Maximum resolution is 1 second.

        Kwargs:

        * only_use_cftime_datetimes (bool):
            If True, will always return cftime datetime objects, regardless of
            calendar.  If False, returns datetime.datetime instances where
            possible.  Defaults to True.

        * only_use_python_datetimes (bool):
            If True, will always return datetime.datetime instances where
            possible, and raise an exception if not.  Ignored if
            only_use_cftime_datetimes is True.  Defaults to False.

        Returns:
            datetime, or numpy.ndarray of datetime object.

        For example:

            >>> import cf_units
            >>> u = cf_units.Unit('hours since 1970-01-01 00:00:00',
            ...                   calendar=cf_units.CALENDAR_STANDARD)
            >>> print(u.num2date(6))
            1970-01-01 06:00:00
            >>> dts = u.num2date([6, 7])
            >>> [str(dt) for dt in dts]
            ['1970-01-01 06:00:00', '1970-01-01 07:00:00']

        """

        return _num2date_to_nearest_second(
            time_value,
            self,
            only_use_cftime_datetimes=only_use_cftime_datetimes,
            only_use_python_datetimes=only_use_python_datetimes,
        )

    def num2pydate(self, time_value):
        """
        Convert time value(s) to python datetime.datetime objects, or raise an
        exception if this is not possible.  Same as::

            unit.num2date(time_value, only_use_cftime_datetimes=False,
                          only_use_python_datetimes=True)

        """
        return self.num2date(
            time_value,
            only_use_cftime_datetimes=False,
            only_use_python_datetimes=True,
        )



# Copyright cf-units contributors
#
# This file is part of cf-units and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

import os
import subprocess
import unittest
from datetime import datetime
from fnmatch import fnmatch
from glob import glob

import cf_units

LICENSE_TEMPLATE = """# Copyright cf-units contributors
#
# This file is part of cf-units and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details."""


# Guess cf_units repo directory of cf_units - realpath is used to mitigate
# against Python finding the cf_units package via a symlink.
DIR = os.path.realpath(os.path.dirname(cf_units.__file__))
REPO_DIR = os.path.dirname(DIR)
DOCS_DIR = os.path.join(REPO_DIR, "doc")
DOCS_DIR = cf_units.config.get_option("Resources", "doc_dir", default=DOCS_DIR)
exclusion = ["Makefile", "make.bat", "build"]
DOCS_DIRS = glob(os.path.join(DOCS_DIR, "*"))
DOCS_DIRS = [
    DOC_DIR
    for DOC_DIR in DOCS_DIRS
    if os.path.basename(DOC_DIR) not in exclusion
]


class TestLicenseHeaders(unittest.TestCase):
    @staticmethod
    def whatchanged_parse(whatchanged_output):
        """
        Returns a generator of tuples of data parsed from
        "git whatchanged --pretty='TIME:%at'". The tuples are of the form
        ``(filename, last_commit_datetime)``

        Sample input::

            ['TIME:1366884020', '',
             ':000000 100644 0000000... 5862ced... A\tcf_units/cf_units.py']

        """
        dt = None
        for line in whatchanged_output:
            if not line.strip():
                continue
            elif line.startswith("TIME:"):
                dt = datetime.fromtimestamp(int(line[5:]))
            else:
                # Non blank, non date, line -> must be the lines
                # containing the file info.
                fname = " ".join(line.split("\t")[1:])
                yield fname, dt

    @staticmethod
    def last_change_by_fname():
        """
        Return a dictionary of all the files under git which maps to
        the datetime of their last modification in the git history.

        .. note::

            This function raises a ValueError if the repo root does
            not have a ".git" folder. If git is not installed on the system,
            or cannot be found by subprocess, an IOError may also be raised.

        """
        # Check the ".git" folder exists at the repo dir.
        if not os.path.isdir(os.path.join(REPO_DIR, ".git")):
            raise ValueError("{} is not a git repository.".format(REPO_DIR))

        # Call "git whatchanged" to get the details of all the files and when
        # they were last changed.
        output = subprocess.check_output(
            ["git", "whatchanged", "--pretty=TIME:%ct"], cwd=REPO_DIR
        )
        output = str(output.decode("ascii"))
        output = output.split("\n")
        res = {}
        for fname, dt in TestLicenseHeaders.whatchanged_parse(output):
            if fname not in res or dt > res[fname]:
                res[fname] = dt

        return res

    def test_license_headers(self):
        exclude_patterns = (
            "setup.py",
            "build/*",
            "dist/*",
            "cf_units/_version.py",
            "cf_units/_udunits2_parser/parser/*",
        )

        try:
            last_change_by_fname = self.last_change_by_fname()
        except ValueError:
            # Caught the case where this is not a git repo.
            return self.skipTest(
                "cf_units installation did not look like a " "git repo."
            )

        failed = False
        for fname, last_change in sorted(last_change_by_fname.items()):
            full_fname = os.path.join(REPO_DIR, fname)
            if (
                full_fname.endswith(".py")
                and os.path.isfile(full_fname)
                and not any(fnmatch(fname, pat) for pat in exclude_patterns)
            ):
                with open(full_fname) as fh:
                    content = fh.read()
                    if not content.startswith(LICENSE_TEMPLATE):
                        print(
                            f"The file {fname} does not start with the "
                            f"required license header."
                        )
                        failed = True

        if failed:
            raise AssertionError(
                "There were license header failures. See stdout."
            )


if __name__ == "__main__":
    unittest.main()



# Copyright cf-units contributors
#
# This file is part of cf-units and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test function :func:`cf_units.num2pydate`."""

import datetime
import unittest

from cf_units import num2pydate


class Test(unittest.TestCase):
    def test_num2pydate_simple(self):
        result = num2pydate(1, "days since 1970-01-01", calendar="standard")
        expected = datetime.datetime(1970, 1, 2)
        self.assertEqual(result, expected)
        self.assertIsInstance(result, datetime.datetime)

    def test_num2pydate_wrong_calendar(self):
        with self.assertRaisesRegex(
            ValueError, "illegal calendar or reference date"
        ):
            num2pydate(1, "days since 1970-01-01", calendar="360_day")


if __name__ == "__main__":
    unittest.main()



# Copyright cf-units contributors
#
# This file is part of cf-units and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.



# Copyright cf-units contributors
#
# This file is part of cf-units and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Miscellaneous utility functions.

"""

import abc
from collections.abc import Hashable


def approx_equal(a, b, max_absolute_error=1e-10, max_relative_error=1e-10):
    """
    Returns whether two numbers are almost equal, allowing for the
    finite precision of floating point numbers.

    """
    # Deal with numbers close to zero
    if abs(a - b) < max_absolute_error:
        return True
    # Ensure we get consistent results if "a" and "b" are supplied in the
    # opposite order.
    max_ab = max([a, b], key=abs)
    relative_error = abs(a - b) / max_ab
    return relative_error < max_relative_error


class _MetaOrderedHashable(abc.ABCMeta):
    """
    A metaclass that ensures that non-abstract subclasses of _OrderedHashable
    without an explicit __init__ method are given a default __init__ method
    with the appropriate method signature.

    Also, an _init method is provided to allow subclasses with their own
    __init__ constructors to initialise their values via an explicit method
    signature.

    NB. This metaclass is used to construct the _OrderedHashable class as well
    as all its subclasses.

    """

    def __new__(cls, name, bases, namespace):
        # We only want to modify concrete classes that have defined the
        # "_names" property.
        if "_names" in namespace and not isinstance(
            namespace["_names"], abc.abstractproperty
        ):
            args = ", ".join(namespace["_names"])

            # Ensure the class has a constructor with explicit arguments.
            if "__init__" not in namespace:
                # Create a default __init__ method for the class
                method_source = (
                    "def __init__(self, %s):\n "
                    "self._init_from_tuple((%s,))" % (args, args)
                )
                exec(method_source, namespace)

            # Ensure the class has a "helper constructor" with explicit
            # arguments.
            if "_init" not in namespace:
                # Create a default _init method for the class
                method_source = (
                    "def _init(self, %s):\n "
                    "self._init_from_tuple((%s,))" % (args, args)
                )
                exec(method_source, namespace)

        return super(_MetaOrderedHashable, cls).__new__(
            cls, name, bases, namespace
        )


class _OrderedHashable(Hashable, metaclass=_MetaOrderedHashable):
    """
    Convenience class for creating "immutable", hashable, and ordered classes.

    Instance identity is defined by the specific list of attribute names
    declared in the abstract attribute "_names". Subclasses must declare the
    attribute "_names" as an iterable containing the names of all the
    attributes relevant to equality/hash-value/ordering.

    Initial values should be set by using ::
        self._init(self, value1, value2, ..)

    .. note::

        It's the responsibility of the subclass to ensure that the values of
        its attributes are themselves hashable.

    """

    pass



# Copyright cf-units contributors
#
# This file is part of cf-units and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `cf_units._udunits2` module.

In most cases, we don't test the correctness
of the operations, only that they return valid objects or raise an
exception where expected."""

import errno
import unittest

import numpy as np

from cf_units import _udunits2 as _ud

_ud.set_error_message_handler(_ud.ignore)


class Test_get_system(unittest.TestCase):
    """
    Test case for operations which create a system object.

    """

    def test_read_xml(self):
        system = _ud.read_xml()

        self.assertIsNotNone(system)

    def test_read_xml_invalid_path(self):
        with self.assertRaises(_ud.UdunitsError) as cm:
            _ud.read_xml(b"/not/a/path.xml")
        ex = cm.exception

        self.assertEqual(ex.status_msg(), "UT_OPEN_ARG")
        self.assertEqual(ex.errnum, errno.ENOENT)


class Test_system(unittest.TestCase):
    """
    Test case for system operations.

    """

    def setUp(self):
        self.system = _ud.read_xml()

    def test_get_unit_by_name(self):
        unit = _ud.get_unit_by_name(self.system, b"metre")

        self.assertIsNotNone(unit)

    def test_get_unit_by_name_invalid_unit(self):
        with self.assertRaises(_ud.UdunitsError):
            _ud.get_unit_by_name(self.system, b"jigawatt")

    def test_parse(self):
        unit = _ud.parse(self.system, b"gigawatt", _ud.UT_ASCII)

        self.assertIsNotNone(unit)

    def test_parse_latin1(self):
        angstrom = _ud.parse(self.system, b"\xe5ngstr\xF6m", _ud.UT_LATIN1)

        self.assertIsNotNone(angstrom)

    def test_parse_ISO_8859_1(self):
        angstrom = _ud.parse(self.system, b"\xe5ngstr\xF6m", _ud.UT_ISO_8859_1)

        self.assertIsNotNone(angstrom)

    def test_parse_UTF8(self):
        angstrom = _ud.parse(
            self.system, b"\xc3\xa5ngstr\xc3\xb6m", _ud.UT_UTF8
        )

        self.assertIsNotNone(angstrom)

    def test_parse_invalid_unit(self):
        with self.assertRaises(_ud.UdunitsError):
            _ud.parse(self.system, b"jigawatt", _ud.UT_ASCII)


class Test_unit(unittest.TestCase):
    """
    Test case for unit operations.

    """

    def setUp(self):
        self.system = _ud.read_xml()
        self.metre = _ud.get_unit_by_name(self.system, b"metre")
        self.yard = _ud.get_unit_by_name(self.system, b"yard")
        self.second = _ud.get_unit_by_name(self.system, b"second")

    def test_clone(self):
        metre_clone = _ud.clone(self.metre)

        self.assertIsNot(metre_clone, self.metre)

    def test_is_dimensionless_true(self):
        radian = _ud.get_unit_by_name(self.system, b"radian")
        self.assertTrue(_ud.is_dimensionless(radian))

    def test_is_dimensionless_false(self):
        self.assertFalse(_ud.is_dimensionless(self.metre))

    def test_compare_same_unit(self):
        self.assertEqual(_ud.compare(self.metre, self.metre), 0)

    def test_compare_diff_unit(self):
        comp = _ud.compare(self.metre, self.second)
        comp_r = _ud.compare(self.second, self.metre)

        self.assertNotEqual(comp, 0)
        self.assertNotEqual(comp_r, 0)
        # m < s iff s > m
        self.assertEqual(comp < 0, comp_r > 0)

    def test_are_convertible_true(self):
        self.assertTrue(_ud.are_convertible(self.metre, self.yard))

    def test_are_convertible_false(self):
        self.assertFalse(_ud.are_convertible(self.metre, self.second))

    def test_get_converter_valid(self):
        _ud.get_converter(self.metre, self.yard)

    def test_get_converter_invalid(self):
        with self.assertRaises(_ud.UdunitsError) as cm:
            _ud.get_converter(self.metre, self.second)
        ex = cm.exception

        self.assertEqual(ex.status_msg(), "UT_MEANINGLESS")

    def test_scale(self):
        mm = _ud.scale(0.001, self.metre)

        self.assertIsNotNone(mm)

    def test_offset(self):
        kelvin = _ud.get_unit_by_name(self.system, b"kelvin")
        celsius = _ud.offset(kelvin, 273.15)

        self.assertIsNotNone(celsius)

    def test_offset_by_time_valid(self):
        time_since = _ud.offset_by_time(self.second, -31622400.0)

        self.assertIsNotNone(time_since)

    def test_offset_by_time_invalid(self):
        with self.assertRaises(_ud.UdunitsError) as cm:
            _ud.offset_by_time(self.metre, -31622400.0)
        cm.exception

        # The udunits package should set a status of UT_MEANINGLESS, according
        # to the documentation. However, it is setting it to UT_SUCCESS.
        # self.assertEqual(ex.status_msg(), 'UT_MEANINGLESS')

    def test_multiply(self):
        metre_second = _ud.multiply(self.metre, self.second)

        self.assertIsNotNone(metre_second)

    def test_invert(self):
        hertz = _ud.invert(self.second)

        self.assertIsNotNone(hertz)

    def test_divide(self):
        metres_per_second = _ud.divide(self.metre, self.second)

        self.assertIsNotNone(metres_per_second)

    def test_raise_(self):
        sq_metre = _ud.raise_(self.metre, 2)

        self.assertIsNotNone(sq_metre)

    def test_root(self):
        hectare = _ud.get_unit_by_name(self.system, b"hectare")
        hundred_metre = _ud.root(hectare, 2)

        self.assertIsNotNone(hundred_metre)

    def test_log(self):
        log_metre = _ud.log(2.7182818, self.metre)

        self.assertIsNotNone(log_metre)

    def test_format(self):
        pascal = _ud.get_unit_by_name(self.system, b"pascal")
        symb = _ud.format(pascal)
        name = _ud.format(pascal, _ud.UT_NAMES)
        defn = _ud.format(pascal, _ud.UT_DEFINITION)
        name_defn = _ud.format(pascal, _ud.UT_DEFINITION | _ud.UT_NAMES)

        self.assertEqual(symb, b"Pa")
        self.assertEqual(name, b"pascal")
        self.assertEqual(defn, b"m-1.kg.s-2")
        self.assertEqual(name_defn, b"meter^-1-kilogram-second^-2")


class Test_time_encoding(unittest.TestCase):
    def setUp(self):
        self.year, self.month, self.day = 2000, 1, 1
        self.date_encoding = -31622400.0
        self.hours, self.minutes, self.seconds = 12, 34, 56
        self.clock_encoding = 45296.0

    def test_encode_date(self):
        res_date_encoding = _ud.encode_date(self.year, self.month, self.day)

        self.assertEqual(self.date_encoding, res_date_encoding)

    def test_encode_clock(self):
        res_clock_encoding = _ud.encode_clock(
            self.hours, self.minutes, self.seconds
        )

        self.assertEqual(self.clock_encoding, res_clock_encoding)

    def test_encode_time(self):
        res_time_encoding = _ud.encode_time(
            self.year,
            self.month,
            self.day,
            self.hours,
            self.minutes,
            self.seconds,
        )

        self.assertEqual(
            self.clock_encoding + self.date_encoding, res_time_encoding
        )

    def test_decode_time(self):
        (
            res_year,
            res_month,
            res_day,
            res_hours,
            res_minutes,
            res_seconds,
            res_resolution,
        ) = _ud.decode_time(self.date_encoding + self.clock_encoding)

        self.assertEqual(
            (res_year, res_month, res_day, res_hours, res_minutes),
            (self.year, self.month, self.day, self.hours, self.minutes),
        )
        self.assertTrue(
            res_seconds - res_resolution
            < self.seconds
            < res_seconds + res_resolution
        )


class Test_convert(unittest.TestCase):
    """
    Test case for convert operations.

    """

    def setUp(self):
        system = _ud.read_xml()
        metre = _ud.get_unit_by_name(system, b"metre")
        yard = _ud.get_unit_by_name(system, b"yard")
        self.converter = _ud.get_converter(metre, yard)
        self.factor = 1.0936132669448853

    def test_convert_float(self):
        res = _ud.convert_float(self.converter, 2.5)
        np.testing.assert_approx_equal(2.5 * self.factor, res)

    def test_convert_floats(self):
        arr = np.array([2.5, 5.0, 10.0], dtype=np.float32)
        res = np.empty_like(arr)
        _ud.convert_floats(self.converter, arr, res)
        np.testing.assert_array_almost_equal(arr * self.factor, res)

    def test_convert_double(self):
        res = _ud.convert_double(self.converter, 2.5)
        np.testing.assert_approx_equal(2.5 * self.factor, res)

    def test_convert_doubles(self):
        arr = np.array([2.5, 5.0, 10.0], dtype=np.float64)
        res = np.empty_like(arr)
        _ud.convert_doubles(self.converter, arr, res)
        np.testing.assert_array_almost_equal(arr * self.factor, res)


if __name__ == "__main__":
    unittest.main()



# Copyright cf-units contributors
#
# This file is part of cf-units and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

"""
Compiles the UDUNITS-2 grammar using ANTLR4.

You may be interested in running this with entr to watch changes to the
grammar:

    echo udunits2*.g4* | entr -c "python compile.py"

You're welcome ;).

"""

import collections
import re
import subprocess
import urllib.request
from pathlib import Path

try:
    import jinja2
except ImportError:
    raise ImportError("Jinja2 needed to compile the grammar.")


JAR_NAME = "antlr-4.7.2-complete.jar"
JAR_URL = f"https://www.antlr.org/download/{JAR_NAME}"
HERE = Path(__file__).resolve().parent

JAR = HERE / JAR_NAME

LEXER = HERE / "parser" / "udunits2Lexer.g4"
PARSER = HERE / "udunits2Parser.g4"


def expand_lexer(source, target):
    MODE_P = re.compile(r"mode ([A-Z_]+)\;")
    TOKEN_P = re.compile(r"([A-Z_]+) ?\:.*")

    with open(source, "r") as fh:
        content = fh.read()

    template = jinja2.Environment(loader=jinja2.BaseLoader).from_string(
        content
    )

    current_mode = "DEFAULT_MODE"

    tokens = collections.defaultdict(list)

    for line in content.split("\n"):
        mode_g = MODE_P.match(line)
        if mode_g:
            current_mode = mode_g.group(1)

        token_g = TOKEN_P.match(line)
        if token_g:
            tokens[current_mode].append(token_g.group(1))

    new_content = template.render(tokens=tokens)
    with open(target, "w") as fh:
        fh.write(new_content)


def main():
    if not JAR.exists():
        print(f"Downloading {JAR_NAME}...")
        urllib.request.urlretrieve(JAR_URL, str(JAR))

    print("Expanding lexer...")
    expand_lexer(LEXER.parent.parent / (LEXER.name + ".jinja"), str(LEXER))

    print("Compiling lexer...")
    subprocess.run(
        [
            "java",
            "-jar",
            str(JAR),
            "-Dlanguage=Python3",
            str(LEXER),
            "-o",
            "parser",
        ],
        check=True,
    )

    print("Compiling parser...")
    subprocess.run(
        [
            "java",
            "-jar",
            str(JAR),
            "-Dlanguage=Python3",
            "-no-listener",
            "-visitor",
            str(PARSER),
            "-o",
            "parser",
        ],
        check=True,
    )

    print("Done.")


if __name__ == "__main__":
    main()



# Copyright cf-units contributors
#
# This file is part of cf-units and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

cdef extern from "udunits2.h":
    ctypedef struct ut_system:
        pass

    ctypedef union ut_unit:
        pass

    ctypedef union cv_converter:
        pass

    ctypedef enum ut_status:
        pass

    ctypedef enum ut_encoding:
        pass

    ctypedef void *ut_error_message_handler

    ut_system* ut_read_xml(char* path)

    void ut_free_system(ut_system* system)

    ut_unit* ut_get_unit_by_name(ut_system* system, char* name)

    ut_unit* ut_clone(ut_unit* unit)

    void ut_free(ut_unit* unit)

    int ut_is_dimensionless(ut_unit* unit)

    int ut_compare(ut_unit* unit1, ut_unit* unit2)

    int ut_are_convertible(ut_unit* unit1, ut_unit* unit2)

    cv_converter* ut_get_converter(ut_unit* fr, ut_unit* to)

    ut_unit* ut_scale(double factor, ut_unit* unit)

    ut_unit* ut_offset(ut_unit* unit, double offset)

    ut_unit* ut_offset_by_time(ut_unit* unit, double origin)

    ut_unit* ut_multiply(ut_unit* unit1, ut_unit* unit2)

    ut_unit* ut_invert(ut_unit* unit)

    ut_unit* ut_divide(ut_unit* numer, ut_unit* denom)

    ut_unit* ut_raise(ut_unit* unit, int power)

    ut_unit* ut_root(ut_unit* unit, int root)

    ut_unit* ut_log(double base, ut_unit* reference)

    ut_unit* ut_parse(ut_system* system, const char* string,
                      ut_encoding encoding)

    int ut_format(ut_unit* unit, char* buf, size_t size, unsigned opts)

    double ut_encode_date(int year, int month, int day)

    double ut_encode_clock(int hours, int minutes, double seconds)

    double ut_encode_time(int year, int month, int day,
                          int hour, int minute, double second)

    void ut_decode_time(double value,
                        int* year, int* month, int* day,
                        int* hour, int* minute, double* second,
                        double* resolution)

    ut_status ut_get_status()

    ut_error_message_handler ut_ignore

    ut_error_message_handler ut_set_error_message_handler(ut_error_message_handler handler)

    float cv_convert_float(cv_converter* converter, float value)

    double cv_convert_double(cv_converter* converter, double value)

    float* cv_convert_floats(cv_converter* converter, float* in_, size_t count, float* out)

    double* cv_convert_doubles(cv_converter* converter, double* const in_, size_t count, double* out)

    void cv_free(cv_converter* conv)




# Copyright cf-units contributors
#
# This file is part of cf-units and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

from cf_units.tex import tex


def test_basic():
    u = "kg kg-1"
    assert tex(u) == r"{kg}\cdot{{kg}^{-1}}"


def test_identifier_micro():
    u = "microW m-2"
    assert tex(u) == r"{{\mu}W}\cdot{{m}^{-2}}"


def test_raise():
    u = "m^2"
    assert tex(u) == r"{m}^{2}"


def test_multiply():
    u = "foo bar"
    assert tex(u) == r"{foo}\cdot{bar}"


def test_divide():
    u = "foo per bar"
    assert tex(u) == r"\frac{foo}{bar}"


def test_shift():
    u = "foo @ 50"
    assert tex(u) == r"{foo} @ {50}"


def test_complex():
    u = "microW^2 / (5^-2)π per sec @ 42"
    expected = r"{\frac{{\frac{{{\mu}W}^{2}}{{5}^{-2}}}\cdot{π}}{sec}} @ {42}"
    assert tex(u) == expected



# Generated from /Users/pelson/dev/scitools/cf-units/cf_units/_udunits2_parser/udunits2Parser.g4 by ANTLR 4.7.2
# encoding: utf-8
import sys
from io import StringIO
from typing.io import TextIO

from antlr4 import *


def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\35")
        buf.write("\u0084\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
        buf.write("\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\3\2\5\2\30\n\2\3\2")
        buf.write("\3\2\3\3\3\3\3\3\5\3\37\n\3\3\3\3\3\5\3#\n\3\3\3\3\3\3")
        buf.write("\3\3\3\5\3)\n\3\3\3\3\3\5\3-\n\3\3\3\3\3\5\3\61\n\3\3")
        buf.write("\4\3\4\3\4\3\4\3\4\3\4\3\4\3\4\3\4\3\4\3\4\3\4\3\4\6\4")
        buf.write("@\n\4\r\4\16\4A\3\4\7\4E\n\4\f\4\16\4H\13\4\3\5\3\5\3")
        buf.write("\5\3\5\3\5\3\5\3\5\3\5\3\5\3\5\3\5\5\5U\n\5\3\6\3\6\3")
        buf.write("\6\3\6\3\6\3\6\5\6]\n\6\3\7\3\7\3\b\3\b\5\bc\n\b\3\t\3")
        buf.write("\t\3\t\5\th\n\t\3\t\3\t\5\tl\n\t\3\t\5\to\n\t\3\t\3\t")
        buf.write("\3\t\5\tt\n\t\3\t\5\tw\n\t\5\ty\n\t\3\n\3\n\3\n\5\n~\n")
        buf.write("\n\3\13\3\13\5\13\u0082\n\13\3\13\2\3\6\f\2\4\6\b\n\f")
        buf.write("\16\20\22\24\2\4\4\2\3\3\13\13\4\2\13\13\33\33\2\u0096")
        buf.write("\2\27\3\2\2\2\4\60\3\2\2\2\6\62\3\2\2\2\bT\3\2\2\2\n\\")
        buf.write("\3\2\2\2\f^\3\2\2\2\16b\3\2\2\2\20x\3\2\2\2\22}\3\2\2")
        buf.write("\2\24\u0081\3\2\2\2\26\30\5\4\3\2\27\26\3\2\2\2\27\30")
        buf.write("\3\2\2\2\30\31\3\2\2\2\31\32\7\2\2\3\32\3\3\2\2\2\33\61")
        buf.write("\5\6\4\2\34\36\5\6\4\2\35\37\7\25\2\2\36\35\3\2\2\2\36")
        buf.write('\37\3\2\2\2\37 \3\2\2\2 "\7\16\2\2!#\7\25\2\2"!\3\2')
        buf.write('\2\2"#\3\2\2\2#$\3\2\2\2$%\5\16\b\2%\61\3\2\2\2&(\5\6')
        buf.write("\4\2')\7\25\2\2('\3\2\2\2()\3\2\2\2)*\3\2\2\2*,\7\16")
        buf.write("\2\2+-\7\25\2\2,+\3\2\2\2,-\3\2\2\2-.\3\2\2\2./\5\20\t")
        buf.write("\2/\61\3\2\2\2\60\33\3\2\2\2\60\34\3\2\2\2\60&\3\2\2\2")
        buf.write("\61\5\3\2\2\2\62\63\b\4\1\2\63\64\5\b\5\2\64F\3\2\2\2")
        buf.write("\65\66\f\6\2\2\66E\5\b\5\2\678\f\5\2\289\7\5\2\29E\5\b")
        buf.write("\5\2:;\f\4\2\2;<\7\6\2\2<E\5\b\5\2=?\f\3\2\2>@\7\25\2")
        buf.write("\2?>\3\2\2\2@A\3\2\2\2A?\3\2\2\2AB\3\2\2\2BC\3\2\2\2C")
        buf.write("E\5\b\5\2D\65\3\2\2\2D\67\3\2\2\2D:\3\2\2\2D=\3\2\2\2")
        buf.write("EH\3\2\2\2FD\3\2\2\2FG\3\2\2\2G\7\3\2\2\2HF\3\2\2\2IJ")
        buf.write("\5\n\6\2JK\5\f\7\2KU\3\2\2\2LU\5\n\6\2MN\5\n\6\2NO\7\20")
        buf.write("\2\2OP\5\f\7\2PU\3\2\2\2QR\5\n\6\2RS\7\17\2\2SU\3\2\2")
        buf.write("\2TI\3\2\2\2TL\3\2\2\2TM\3\2\2\2TQ\3\2\2\2U\t\3\2\2\2")
        buf.write("V]\7\23\2\2WX\7\b\2\2XY\5\4\3\2YZ\7\t\2\2Z]\3\2\2\2[]")
        buf.write("\5\16\b\2\\V\3\2\2\2\\W\3\2\2\2\\[\3\2\2\2]\13\3\2\2\2")
        buf.write("^_\t\2\2\2_\r\3\2\2\2`c\5\f\7\2ac\7\r\2\2b`\3\2\2\2ba")
        buf.write("\3\2\2\2c\17\3\2\2\2dy\t\3\2\2eg\t\3\2\2fh\7\25\2\2gf")
        buf.write("\3\2\2\2gh\3\2\2\2hi\3\2\2\2in\5\22\n\2jl\7\25\2\2kj\3")
        buf.write("\2\2\2kl\3\2\2\2lm\3\2\2\2mo\5\24\13\2nk\3\2\2\2no\3\2")
        buf.write("\2\2oy\3\2\2\2py\7\35\2\2qs\7\34\2\2rt\7\25\2\2sr\3\2")
        buf.write("\2\2st\3\2\2\2tv\3\2\2\2uw\5\24\13\2vu\3\2\2\2vw\3\2\2")
        buf.write("\2wy\3\2\2\2xd\3\2\2\2xe\3\2\2\2xp\3\2\2\2xq\3\2\2\2y")
        buf.write("\21\3\2\2\2z~\7\30\2\2{~\7\31\2\2|~\5\f\7\2}z\3\2\2\2")
        buf.write("}{\3\2\2\2}|\3\2\2\2~\23\3\2\2\2\177\u0082\7\31\2\2\u0080")
        buf.write("\u0082\5\f\7\2\u0081\177\3\2\2\2\u0081\u0080\3\2\2\2\u0082")
        buf.write('\25\3\2\2\2\26\27\36"(,\60ADFT\\bgknsvx}\u0081')
        return buf.getvalue()


class udunits2Parser(Parser):

    grammarFileName = "udunits2Parser.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [DFA(ds, i) for i, ds in enumerate(atn.decisionToState)]

    sharedContextCache = PredictionContextCache()

    literalNames = [
        "<INVALID>",
        "<INVALID>",
        "'+'",
        "<INVALID>",
        "<INVALID>",
        "'.'",
        "'('",
        "')'",
        "':'",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "'-'",
    ]

    symbolicNames = [
        "<INVALID>",
        "SIGNED_INT",
        "PLUS",
        "MULTIPLY",
        "DIVIDE",
        "PERIOD",
        "OPEN_PAREN",
        "CLOSE_PAREN",
        "SEMICOLON",
        "INT",
        "E_POWER",
        "FLOAT",
        "SHIFT_OP",
        "UNICODE_EXPONENT",
        "RAISE",
        "LOG",
        "LOGREF",
        "ID",
        "LATIN_SUBSET",
        "WS",
        "ERRORCHARACTER",
        "TIMEZONE",
        "HOUR_MINUTE_SECOND",
        "HOUR_MINUTE",
        "M_MINUS",
        "DATE",
        "TIMESTAMP",
        "DT_T_CLOCK",
    ]

    RULE_unit_spec = 0
    RULE_shift_spec = 1
    RULE_product = 2
    RULE_power = 3
    RULE_basic_spec = 4
    RULE_integer = 5
    RULE_number = 6
    RULE_timestamp = 7
    RULE_signed_clock = 8
    RULE_timezone_offset = 9

    ruleNames = [
        "unit_spec",
        "shift_spec",
        "product",
        "power",
        "basic_spec",
        "integer",
        "number",
        "timestamp",
        "signed_clock",
        "timezone_offset",
    ]

    EOF = Token.EOF
    SIGNED_INT = 1
    PLUS = 2
    MULTIPLY = 3
    DIVIDE = 4
    PERIOD = 5
    OPEN_PAREN = 6
    CLOSE_PAREN = 7
    SEMICOLON = 8
    INT = 9
    E_POWER = 10
    FLOAT = 11
    SHIFT_OP = 12
    UNICODE_EXPONENT = 13
    RAISE = 14
    LOG = 15
    LOGREF = 16
    ID = 17
    LATIN_SUBSET = 18
    WS = 19
    ERRORCHARACTER = 20
    TIMEZONE = 21
    HOUR_MINUTE_SECOND = 22
    HOUR_MINUTE = 23
    M_MINUS = 24
    DATE = 25
    TIMESTAMP = 26
    DT_T_CLOCK = 27

    def __init__(self, input: TokenStream, output: TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.7.2")
        self._interp = ParserATNSimulator(
            self, self.atn, self.decisionsToDFA, self.sharedContextCache
        )
        self._predicates = None

    class Unit_specContext(ParserRuleContext):
        def __init__(
            self,
            parser,
            parent: ParserRuleContext = None,
            invokingState: int = -1,
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(udunits2Parser.EOF, 0)

        def shift_spec(self):
            return self.getTypedRuleContext(
                udunits2Parser.Shift_specContext, 0
            )

        def getRuleIndex(self):
            return udunits2Parser.RULE_unit_spec

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitUnit_spec"):
                return visitor.visitUnit_spec(self)
            else:
                return visitor.visitChildren(self)

    def unit_spec(self):

        localctx = udunits2Parser.Unit_specContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_unit_spec)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 21
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if ((_la) & ~0x3F) == 0 and (
                (1 << _la)
                & (
                    (1 << udunits2Parser.SIGNED_INT)
                    | (1 << udunits2Parser.OPEN_PAREN)
                    | (1 << udunits2Parser.INT)
                    | (1 << udunits2Parser.FLOAT)
                    | (1 << udunits2Parser.ID)
                )
            ) != 0:
                self.state = 20
                self.shift_spec()

            self.state = 23
            self.match(udunits2Parser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Shift_specContext(ParserRuleContext):
        def __init__(
            self,
            parser,
            parent: ParserRuleContext = None,
            invokingState: int = -1,
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def product(self):
            return self.getTypedRuleContext(udunits2Parser.ProductContext, 0)

        def SHIFT_OP(self):
            return self.getToken(udunits2Parser.SHIFT_OP, 0)

        def number(self):
            return self.getTypedRuleContext(udunits2Parser.NumberContext, 0)

        def WS(self, i: int = None):
            if i is None:
                return self.getTokens(udunits2Parser.WS)
            else:
                return self.getToken(udunits2Parser.WS, i)

        def timestamp(self):
            return self.getTypedRuleContext(udunits2Parser.TimestampContext, 0)

        def getRuleIndex(self):
            return udunits2Parser.RULE_shift_spec

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitShift_spec"):
                return visitor.visitShift_spec(self)
            else:
                return visitor.visitChildren(self)

    def shift_spec(self):

        localctx = udunits2Parser.Shift_specContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 2, self.RULE_shift_spec)
        self._la = 0  # Token type
        try:
            self.state = 46
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 5, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 25
                self.product(0)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 26
                self.product(0)
                self.state = 28
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == udunits2Parser.WS:
                    self.state = 27
                    self.match(udunits2Parser.WS)

                self.state = 30
                self.match(udunits2Parser.SHIFT_OP)
                self.state = 32
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == udunits2Parser.WS:
                    self.state = 31
                    self.match(udunits2Parser.WS)

                self.state = 34
                self.number()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 36
                self.product(0)
                self.state = 38
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == udunits2Parser.WS:
                    self.state = 37
                    self.match(udunits2Parser.WS)

                self.state = 40
                self.match(udunits2Parser.SHIFT_OP)
                self.state = 42
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == udunits2Parser.WS:
                    self.state = 41
                    self.match(udunits2Parser.WS)

                self.state = 44
                self.timestamp()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ProductContext(ParserRuleContext):
        def __init__(
            self,
            parser,
            parent: ParserRuleContext = None,
            invokingState: int = -1,
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def power(self):
            return self.getTypedRuleContext(udunits2Parser.PowerContext, 0)

        def product(self):
            return self.getTypedRuleContext(udunits2Parser.ProductContext, 0)

        def MULTIPLY(self):
            return self.getToken(udunits2Parser.MULTIPLY, 0)

        def DIVIDE(self):
            return self.getToken(udunits2Parser.DIVIDE, 0)

        def WS(self, i: int = None):
            if i is None:
                return self.getTokens(udunits2Parser.WS)
            else:
                return self.getToken(udunits2Parser.WS, i)

        def getRuleIndex(self):
            return udunits2Parser.RULE_product

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitProduct"):
                return visitor.visitProduct(self)
            else:
                return visitor.visitChildren(self)

    def product(self, _p: int = 0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = udunits2Parser.ProductContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 4
        self.enterRecursionRule(localctx, 4, self.RULE_product, _p)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 49
            self.power()
            self._ctx.stop = self._input.LT(-1)
            self.state = 68
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 8, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 66
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(
                        self._input, 7, self._ctx
                    )
                    if la_ == 1:
                        localctx = udunits2Parser.ProductContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_product
                        )
                        self.state = 51
                        if not self.precpred(self._ctx, 4):
                            from antlr4.error.Errors import (
                                FailedPredicateException,
                            )

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 4)"
                            )
                        self.state = 52
                        self.power()
                        pass

                    elif la_ == 2:
                        localctx = udunits2Parser.ProductContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_product
                        )
                        self.state = 53
                        if not self.precpred(self._ctx, 3):
                            from antlr4.error.Errors import (
                                FailedPredicateException,
                            )

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 3)"
                            )
                        self.state = 54
                        self.match(udunits2Parser.MULTIPLY)
                        self.state = 55
                        self.power()
                        pass

                    elif la_ == 3:
                        localctx = udunits2Parser.ProductContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_product
                        )
                        self.state = 56
                        if not self.precpred(self._ctx, 2):
                            from antlr4.error.Errors import (
                                FailedPredicateException,
                            )

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 2)"
                            )
                        self.state = 57
                        self.match(udunits2Parser.DIVIDE)
                        self.state = 58
                        self.power()
                        pass

                    elif la_ == 4:
                        localctx = udunits2Parser.ProductContext(
                            self, _parentctx, _parentState
                        )
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_product
                        )
                        self.state = 59
                        if not self.precpred(self._ctx, 1):
                            from antlr4.error.Errors import (
                                FailedPredicateException,
                            )

                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 1)"
                            )
                        self.state = 61
                        self._errHandler.sync(self)
                        _la = self._input.LA(1)
                        while True:
                            self.state = 60
                            self.match(udunits2Parser.WS)
                            self.state = 63
                            self._errHandler.sync(self)
                            _la = self._input.LA(1)
                            if not (_la == udunits2Parser.WS):
                                break

                        self.state = 65
                        self.power()
                        pass

                self.state = 70
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 8, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class PowerContext(ParserRuleContext):
        def __init__(
            self,
            parser,
            parent: ParserRuleContext = None,
            invokingState: int = -1,
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def basic_spec(self):
            return self.getTypedRuleContext(
                udunits2Parser.Basic_specContext, 0
            )

        def integer(self):
            return self.getTypedRuleContext(udunits2Parser.IntegerContext, 0)

        def RAISE(self):
            return self.getToken(udunits2Parser.RAISE, 0)

        def UNICODE_EXPONENT(self):
            return self.getToken(udunits2Parser.UNICODE_EXPONENT, 0)

        def getRuleIndex(self):
            return udunits2Parser.RULE_power

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitPower"):
                return visitor.visitPower(self)
            else:
                return visitor.visitChildren(self)

    def power(self):

        localctx = udunits2Parser.PowerContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_power)
        try:
            self.state = 82
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 9, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 71
                self.basic_spec()
                self.state = 72
                self.integer()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 74
                self.basic_spec()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 75
                self.basic_spec()
                self.state = 76
                self.match(udunits2Parser.RAISE)
                self.state = 77
                self.integer()
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 79
                self.basic_spec()
                self.state = 80
                self.match(udunits2Parser.UNICODE_EXPONENT)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Basic_specContext(ParserRuleContext):
        def __init__(
            self,
            parser,
            parent: ParserRuleContext = None,
            invokingState: int = -1,
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(udunits2Parser.ID, 0)

        def OPEN_PAREN(self):
            return self.getToken(udunits2Parser.OPEN_PAREN, 0)

        def shift_spec(self):
            return self.getTypedRuleContext(
                udunits2Parser.Shift_specContext, 0
            )

        def CLOSE_PAREN(self):
            return self.getToken(udunits2Parser.CLOSE_PAREN, 0)

        def number(self):
            return self.getTypedRuleContext(udunits2Parser.NumberContext, 0)

        def getRuleIndex(self):
            return udunits2Parser.RULE_basic_spec

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitBasic_spec"):
                return visitor.visitBasic_spec(self)
            else:
                return visitor.visitChildren(self)

    def basic_spec(self):

        localctx = udunits2Parser.Basic_specContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 8, self.RULE_basic_spec)
        try:
            self.state = 90
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [udunits2Parser.ID]:
                self.enterOuterAlt(localctx, 1)
                self.state = 84
                self.match(udunits2Parser.ID)
                pass
            elif token in [udunits2Parser.OPEN_PAREN]:
                self.enterOuterAlt(localctx, 2)
                self.state = 85
                self.match(udunits2Parser.OPEN_PAREN)
                self.state = 86
                self.shift_spec()
                self.state = 87
                self.match(udunits2Parser.CLOSE_PAREN)
                pass
            elif token in [
                udunits2Parser.SIGNED_INT,
                udunits2Parser.INT,
                udunits2Parser.FLOAT,
            ]:
                self.enterOuterAlt(localctx, 3)
                self.state = 89
                self.number()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class IntegerContext(ParserRuleContext):
        def __init__(
            self,
            parser,
            parent: ParserRuleContext = None,
            invokingState: int = -1,
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def INT(self):
            return self.getToken(udunits2Parser.INT, 0)

        def SIGNED_INT(self):
            return self.getToken(udunits2Parser.SIGNED_INT, 0)

        def getRuleIndex(self):
            return udunits2Parser.RULE_integer

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitInteger"):
                return visitor.visitInteger(self)
            else:
                return visitor.visitChildren(self)

    def integer(self):

        localctx = udunits2Parser.IntegerContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_integer)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 92
            _la = self._input.LA(1)
            if not (
                _la == udunits2Parser.SIGNED_INT or _la == udunits2Parser.INT
            ):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class NumberContext(ParserRuleContext):
        def __init__(
            self,
            parser,
            parent: ParserRuleContext = None,
            invokingState: int = -1,
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def integer(self):
            return self.getTypedRuleContext(udunits2Parser.IntegerContext, 0)

        def FLOAT(self):
            return self.getToken(udunits2Parser.FLOAT, 0)

        def getRuleIndex(self):
            return udunits2Parser.RULE_number

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitNumber"):
                return visitor.visitNumber(self)
            else:
                return visitor.visitChildren(self)

    def number(self):

        localctx = udunits2Parser.NumberContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_number)
        try:
            self.state = 96
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [udunits2Parser.SIGNED_INT, udunits2Parser.INT]:
                self.enterOuterAlt(localctx, 1)
                self.state = 94
                self.integer()
                pass
            elif token in [udunits2Parser.FLOAT]:
                self.enterOuterAlt(localctx, 2)
                self.state = 95
                self.match(udunits2Parser.FLOAT)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class TimestampContext(ParserRuleContext):
        def __init__(
            self,
            parser,
            parent: ParserRuleContext = None,
            invokingState: int = -1,
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DATE(self):
            return self.getToken(udunits2Parser.DATE, 0)

        def INT(self):
            return self.getToken(udunits2Parser.INT, 0)

        def signed_clock(self):
            return self.getTypedRuleContext(
                udunits2Parser.Signed_clockContext, 0
            )

        def WS(self, i: int = None):
            if i is None:
                return self.getTokens(udunits2Parser.WS)
            else:
                return self.getToken(udunits2Parser.WS, i)

        def timezone_offset(self):
            return self.getTypedRuleContext(
                udunits2Parser.Timezone_offsetContext, 0
            )

        def DT_T_CLOCK(self):
            return self.getToken(udunits2Parser.DT_T_CLOCK, 0)

        def TIMESTAMP(self):
            return self.getToken(udunits2Parser.TIMESTAMP, 0)

        def getRuleIndex(self):
            return udunits2Parser.RULE_timestamp

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitTimestamp"):
                return visitor.visitTimestamp(self)
            else:
                return visitor.visitChildren(self)

    def timestamp(self):

        localctx = udunits2Parser.TimestampContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_timestamp)
        self._la = 0  # Token type
        try:
            self.state = 118
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 17, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 98
                _la = self._input.LA(1)
                if not (
                    _la == udunits2Parser.INT or _la == udunits2Parser.DATE
                ):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 99
                _la = self._input.LA(1)
                if not (
                    _la == udunits2Parser.INT or _la == udunits2Parser.DATE
                ):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 101
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == udunits2Parser.WS:
                    self.state = 100
                    self.match(udunits2Parser.WS)

                self.state = 103
                self.signed_clock()
                self.state = 108
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if ((_la) & ~0x3F) == 0 and (
                    (1 << _la)
                    & (
                        (1 << udunits2Parser.SIGNED_INT)
                        | (1 << udunits2Parser.INT)
                        | (1 << udunits2Parser.WS)
                        | (1 << udunits2Parser.HOUR_MINUTE)
                    )
                ) != 0:
                    self.state = 105
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if _la == udunits2Parser.WS:
                        self.state = 104
                        self.match(udunits2Parser.WS)

                    self.state = 107
                    self.timezone_offset()

                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 110
                self.match(udunits2Parser.DT_T_CLOCK)
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 111
                self.match(udunits2Parser.TIMESTAMP)
                self.state = 113
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == udunits2Parser.WS:
                    self.state = 112
                    self.match(udunits2Parser.WS)

                self.state = 116
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if ((_la) & ~0x3F) == 0 and (
                    (1 << _la)
                    & (
                        (1 << udunits2Parser.SIGNED_INT)
                        | (1 << udunits2Parser.INT)
                        | (1 << udunits2Parser.HOUR_MINUTE)
                    )
                ) != 0:
                    self.state = 115
                    self.timezone_offset()

                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Signed_clockContext(ParserRuleContext):
        def __init__(
            self,
            parser,
            parent: ParserRuleContext = None,
            invokingState: int = -1,
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def HOUR_MINUTE_SECOND(self):
            return self.getToken(udunits2Parser.HOUR_MINUTE_SECOND, 0)

        def HOUR_MINUTE(self):
            return self.getToken(udunits2Parser.HOUR_MINUTE, 0)

        def integer(self):
            return self.getTypedRuleContext(udunits2Parser.IntegerContext, 0)

        def getRuleIndex(self):
            return udunits2Parser.RULE_signed_clock

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSigned_clock"):
                return visitor.visitSigned_clock(self)
            else:
                return visitor.visitChildren(self)

    def signed_clock(self):

        localctx = udunits2Parser.Signed_clockContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 16, self.RULE_signed_clock)
        try:
            self.state = 123
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [udunits2Parser.HOUR_MINUTE_SECOND]:
                self.enterOuterAlt(localctx, 1)
                self.state = 120
                self.match(udunits2Parser.HOUR_MINUTE_SECOND)
                pass
            elif token in [udunits2Parser.HOUR_MINUTE]:
                self.enterOuterAlt(localctx, 2)
                self.state = 121
                self.match(udunits2Parser.HOUR_MINUTE)
                pass
            elif token in [udunits2Parser.SIGNED_INT, udunits2Parser.INT]:
                self.enterOuterAlt(localctx, 3)
                self.state = 122
                self.integer()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Timezone_offsetContext(ParserRuleContext):
        def __init__(
            self,
            parser,
            parent: ParserRuleContext = None,
            invokingState: int = -1,
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def HOUR_MINUTE(self):
            return self.getToken(udunits2Parser.HOUR_MINUTE, 0)

        def integer(self):
            return self.getTypedRuleContext(udunits2Parser.IntegerContext, 0)

        def getRuleIndex(self):
            return udunits2Parser.RULE_timezone_offset

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitTimezone_offset"):
                return visitor.visitTimezone_offset(self)
            else:
                return visitor.visitChildren(self)

    def timezone_offset(self):

        localctx = udunits2Parser.Timezone_offsetContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 18, self.RULE_timezone_offset)
        try:
            self.state = 127
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [udunits2Parser.HOUR_MINUTE]:
                self.enterOuterAlt(localctx, 1)
                self.state = 125
                self.match(udunits2Parser.HOUR_MINUTE)
                pass
            elif token in [udunits2Parser.SIGNED_INT, udunits2Parser.INT]:
                self.enterOuterAlt(localctx, 2)
                self.state = 126
                self.integer()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    def sempred(self, localctx: RuleContext, ruleIndex: int, predIndex: int):
        if self._predicates == None:
            self._predicates = dict()
        self._predicates[2] = self.product_sempred
        pred = self._predicates.get(ruleIndex, None)
        if pred is None:
            raise Exception("No predicate with index:" + str(ruleIndex))
        else:
            return pred(localctx, predIndex)

    def product_sempred(self, localctx: ProductContext, predIndex: int):
        if predIndex == 0:
            return self.precpred(self._ctx, 4)

        if predIndex == 1:
            return self.precpred(self._ctx, 3)

        if predIndex == 2:
            return self.precpred(self._ctx, 2)

        if predIndex == 3:
            return self.precpred(self._ctx, 1)
